from typing import List, Tuple

import numpy as np
import pytest
import torch
from scipy.spatial.distance import squareform
from torch import Tensor, tensor

from oml.miners.inbatch_hard_tri import HardTripletsMiner
from oml.utils.misc import find_value_ids
from tests.test_oml.test_miners.conftest import generate_valid_labels
from tests.test_oml.test_miners.shared_checkers import check_triplets_consistency


@pytest.fixture()
def distmats_and_labels() -> List[Tuple[Tensor, List[int]]]:
    """
    Returns: list of distance matrices and valid labels

    """
    num_batches = 100

    labels_li = generate_valid_labels(num=num_batches)
    labels_list, _, _ = zip(*labels_li)

    distmats = []
    for labels in labels_list:
        n = len(labels)
        distmats.append(tensor(squareform(torch.rand(int(n * (n - 1) / 2)))))

    return list(zip(distmats, labels_list))


def test_hard_miner_from_features(features_and_labels) -> None:  # type: ignore
    """
    Args:
        features_and_labels: Features and valid labels

    """
    miner = HardTripletsMiner(norm_required=True)

    for features, labels in features_and_labels:
        ids_a, ids_p, ids_n = miner._sample(features=features, labels=labels)

        check_triplets_consistency(ids_anchor=ids_a, ids_pos=ids_p, ids_neg=ids_n, labels=labels)

        assert len(ids_a) == len(labels)


def test_hard_miner_from_dist(distmats_and_labels) -> None:  # type: ignore
    """
    Args:
        distmats_and_labels:
            List of distance matrices and valid labels

    """
    miner = HardTripletsMiner(norm_required=True)

    for distmat, labels in distmats_and_labels:
        ids_a, ids_p, ids_n = miner._sample_from_distmat(distmat=distmat, labels=labels)

        check_triplets_are_hardest(
            ids_anchor=ids_a,
            ids_pos=ids_p,
            ids_neg=ids_n,
            labels=labels,
            distmat=distmat,
        )

        check_triplets_consistency(ids_anchor=ids_a, ids_pos=ids_p, ids_neg=ids_n, labels=labels)

        assert len(labels) == len(ids_a)


def test_hard_miner_manual() -> None:
    """
    Test on manual example.

    """
    labels = [0, 0, 1, 1]

    dist_mat = torch.tensor(
        [
            [0.0, 0.3, 0.2, 0.4],
            [0.3, 0.0, 0.4, 0.8],
            [0.2, 0.4, 0.0, 0.5],
            [0.4, 0.8, 0.5, 0.0],
        ]
    )

    gt = {(0, 1, 2), (1, 0, 2), (2, 3, 0), (3, 2, 0)}

    miner = HardTripletsMiner(norm_required=True)

    ids_a, ids_p, ids_n = miner._sample_from_distmat(distmat=dist_mat, labels=labels)
    predict = set(zip(ids_a, ids_p, ids_n))

    check_triplets_consistency(ids_anchor=ids_a, ids_pos=ids_p, ids_neg=ids_n, labels=labels)

    assert len(labels) == len(ids_a)
    assert predict == gt


def check_triplets_are_hardest(
    ids_anchor: List[int],
    ids_pos: List[int],
    ids_neg: List[int],
    labels: List[int],
    distmat: Tensor,
) -> None:
    """
    Args:
        ids_anchor: Anchor indexes of selected triplets
        ids_pos: Positive indexes of selected triplets
        ids_neg: Negative indexes of selected triplets
        labels: Labels of the samples in the batch
        distmat: Distances between features

    """
    ids_all = set(range(len(labels)))

    for i_a, i_p, i_n in zip(ids_anchor, ids_pos, ids_neg):
        ids_label = set(find_value_ids(it=labels, value=labels[i_a]))

        ids_pos_cur = np.array(list(ids_label - {i_a}), int)
        ids_neg_cur = np.array(list(ids_all - ids_label), int)

        assert torch.isclose(distmat[i_a, ids_pos_cur].max(), distmat[i_a, i_p])

        assert torch.isclose(distmat[i_a, ids_neg_cur].min(), distmat[i_a, i_n])



from abc import ABC

import numpy as np
from torch import Tensor, nn


class IExtractor(nn.Module, ABC):
    """
    Models have to inherit this interface to be comparable with the rest of the library.

    """

    def extract(self, x: Tensor) -> Tensor:
        return self.forward(x)

    @property
    def feat_dim(self) -> int:
        raise NotImplementedError()


__all__ = ["IExtractor"]



import pytest
import torch
from torch.nn import TripletMarginLoss

from oml.losses.triplet import TripletLoss


@pytest.mark.parametrize("reduction", ["sum", "mean", "none"])
@pytest.mark.parametrize("margin", [0.2, 0.5, 1.5])
def test_triplet_loss_vs_torch_version(reduction: str, margin: float) -> None:
    criterion_torch = TripletMarginLoss(reduction=reduction, margin=margin)
    criterion_our = TripletLoss(reduction=reduction, margin=margin, need_logs=True)

    anchor = 1 * torch.ones(32, 1024)
    positive = 2 * torch.ones(32, 1024)
    negative = 3 * torch.ones(32, 1024)

    x_torch = criterion_torch(anchor, positive, negative)
    x_our = criterion_our(anchor, positive, negative)

    assert x_torch.allclose(x_our, rtol=0.01)


def test_soft_triplet_loss() -> None:
    vec = torch.randn(32, 1024)
    gt = torch.tensor(0.693147180559945)  # this value is log1p(0)

    assert TripletLoss(margin=None, reduction="mean", need_logs=True)(vec, vec, vec).isclose(gt)



from typing import Any, Dict

import numpy as np
import pytest
import torch

from oml.functional.metrics import reduce_metrics


@pytest.mark.parametrize(
    "metrics_dict,result_dict",
    [
        ({"OVERALL": {"cmc": {1: torch.tensor([0.1, 0.2, 0.0])}}}, {"OVERALL": {"cmc": {1: 0.1}}}),
        ({"cmc": {1: 1}}, {"cmc": {1: 1}}),
        ({"OVERALL": {"cmc": {1: torch.tensor([0.1, 0.2, 0.0])}}, "a": 0.3}, {"OVERALL": {"cmc": {1: 0.1}}, "a": 0.3}),
        (
            {"OVERALL": {"cmc": {1: torch.tensor([0.1, 0.2, 0.0])}}, "a": torch.tensor([0.3, 0.1])},
            {"OVERALL": {"cmc": {1: 0.1}}, "a": 0.2},
        ),
        (
            {"OVERALL": {"cmc": {1: np.r_[0.1, 0.2, 0.0]}}, "a": np.r_[0.3, 0.1]},
            {"OVERALL": {"cmc": {1: np.mean(np.r_[0.1, 0.2, 0.0])}}, "a": np.mean(np.r_[0.3, 0.1])},
        ),
        (
            {"OVERALL": {"cmc": {1: torch.tensor([0.1, 0.2, 0.0])}}, "a": torch.tensor([0.3])},
            {"OVERALL": {"cmc": {1: 0.1}}, "a": 0.3},
        ),
        pytest.param({"OVERALL": {"cmc": {1: {}}}}, {"OVERALL": {"cmc": {}}}, marks=pytest.mark.xfail),
    ],
)
def test_reduce_metrics(metrics_dict: Dict[str, Any], result_dict: Dict[str, Any]) -> None:
    assert reduce_metrics(metrics_dict) == result_dict, reduce_metrics(metrics_dict)



from pathlib import Path
from typing import Optional, Union

import numpy as np
import torch
from pytorch_grad_cam import GradCAM
from pytorch_grad_cam.utils.image import show_cam_on_image
from torch import nn
from torchvision.models import resnet18, resnet34, resnet50, resnet101, resnet152

from oml.interfaces.models import IExtractor
from oml.models.pooling import GEM
from oml.models.utils import remove_prefix_from_state_dict
from oml.transforms.images.albumentations.transforms import get_normalisation_albu
from oml.utils.io import download_checkpoint


class ResnetExtractor(IExtractor):
    """
    The base class for the extractors that follow ResNet architecture.

    """

    constructors = {
        "resnet18": resnet18,
        "resnet34": resnet34,
        "resnet50": resnet50,
        "resnet101": resnet101,
        "resnet152": resnet152,
    }

    pretrained_models = {
        "resnet50_moco_v2": (
            "https://dl.fbaipublicfiles.com/moco/moco_checkpoints/moco_v2_800ep/moco_v2_800ep_pretrain.pth.tar",
            "a04e12f8",
            None,
        )
    }

    def __init__(
        self,
        weights: Optional[Union[Path, str]],
        arch: str,
        normalise_features: bool,
        gem_p: Optional[float],
        remove_fc: bool,
        strict_load: bool = True,
    ):
        """

        Args:
            weights: Path to weights or a special key to download pretrained checkpoint, use ``None`` to randomly initialize
             model's weights or ``default`` to use the checkpoint pretrained on ImageNet.
             You can check the available pretrained checkpoints in ``self.pretrained_models``.
            arch: Different types of ResNet, please, check ``self.constructors``
            normalise_features: Set ``True`` to normalise output features
            gem_p: Value of power in `Generalized Mean Pooling` that we use as the replacement for the default one
             (if ``gem_p == 1`` it's just a normal average pooling and if ``gem_p -> inf`` it's max-pooling)
            remove_fc: Set ``True`` if you want to remove the last fully connected layer
            strict_load: Set ``True`` if you want the strict load of the weights from the checkpoint

        """
        assert arch in self.constructors.keys()

        super(ResnetExtractor, self).__init__()

        self.arch = arch
        self.normalise_features = normalise_features
        self.remove_fc = remove_fc

        factory_fun = self.constructors[self.arch]

        self.model = factory_fun(weights=None)

        if weights == "resnet50_moco_v2":
            # todo:
            # in the next iterations we should use head as a separate entity to change fc
            self.model.fc = nn.Sequential(
                nn.Linear(self.model.fc.weight.shape[1], self.model.fc.weight.shape[1]),
                nn.ReLU(),
                nn.Linear(self.model.fc.weight.shape[1], 128),  # output size in moco v2
            )

        self.last_conv_channels = self.calc_last_conv_channels()

        if gem_p is not None:
            self.model.avgpool = GEM(p=gem_p)

        if weights is None:
            return

        elif isinstance(weights, str) and weights.lower() == "default":
            state_dict = factory_fun(weights="DEFAULT").state_dict()

        elif weights in self.pretrained_models:
            url_or_fid, hash_md5, fname = self.pretrained_models[weights]  # type: ignore
            path_to_ckpt = download_checkpoint(url_or_fid=url_or_fid, hash_md5=hash_md5, fname=fname)
            state_dict = load_moco_model(path_to_model=Path(path_to_ckpt)).state_dict()

        else:
            state_dict = torch.load(weights, map_location="cpu")

        state_dict = state_dict["state_dict"] if "state_dict" in state_dict.keys() else state_dict
        state_dict = remove_prefix_from_state_dict(state_dict, "layer4.")
        self.model.load_state_dict(state_dict, strict=strict_load)

        if remove_fc:
            self.model.fc = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)

        if self.normalise_features:
            xn = torch.linalg.norm(x, 2, dim=1).detach()
            x = x.div(xn.unsqueeze(1))

        return x

    def calc_last_conv_channels(self) -> int:
        last_block = self.model.layer4[-1]
        if self.arch in ("resnet18", "resnet34"):
            n_out_channels = last_block.conv2.out_channels
        else:
            # resnet50, resnet101, resnet152
            n_out_channels = last_block.conv3.out_channels
        return n_out_channels

    @property
    def feat_dim(self) -> int:
        if isinstance(self.model.fc, torch.nn.Identity):
            return self.last_conv_channels
        elif isinstance(self.model.fc, torch.nn.Linear):
            return self.model.fc.out_features
        else:
            # 2-layer mlp case
            return self.model.fc[-1].out_features

    def draw_gradcam(self, image: np.ndarray) -> np.ndarray:
        """
        Visualization of the gradients on a particular image using `GradCam`_.

        .. _GradCam: https://arxiv.org/abs/1610.02391

        """
        model_device = str(list(self.model.parameters())[0].device)
        image_tensor = get_normalisation_albu()(image=image)["image"].to(model_device)
        cam = GradCAM(model=self.model, target_layer=self.model.layer4[-1], use_cuda=not (model_device == "cpu"))
        gray_image = cam(image_tensor.unsqueeze(0), "gradcam", None)
        img_with_grads = show_cam_on_image(image / 255, gray_image)
        return img_with_grads


def load_moco_model(path_to_model: Path) -> nn.Module:
    """
    Args:
        path_to_model: Path to model trained using original
           code from MoCo repository:
           https://github.com/facebookresearch/moco

    Returns:
        Model

    """
    checkpoint = torch.load(path_to_model, map_location="cpu")

    state_dict = checkpoint["state_dict"]
    for k in list(state_dict.keys()):
        # retain only encoder_q up to before the embedding layer
        if k.startswith("module.encoder_q"):
            state_dict[k[len("module.encoder_q.") :]] = state_dict[k]
        del state_dict[k]

    model = resnet50(num_classes=128)  # output size in moco v2

    if "fc.2.weight" in state_dict.keys():
        print("MOCO V2 architecture was detected!")
        model.fc = nn.Sequential(
            nn.Linear(model.fc.weight.shape[1], model.fc.weight.shape[1]),
            nn.ReLU(),
            model.fc,
        )
    else:
        print("MOCO V1 architecture will be used")

    model.load_state_dict(state_dict)

    return model


__all__ = ["ResnetExtractor", "load_moco_model"]



import warnings
from collections import defaultdict
from typing import Dict, Optional, Tuple, Union

import numpy as np
import torch

from oml.losses.triplet import get_tri_ids_in_plain
from oml.utils.misc import clip_max
from oml.utils.misc_torch import elementwise_dist, pairwise_dist

TMetricsDict = Dict[str, Dict[int, Union[float, torch.Tensor]]]


def calc_retrieval_metrics(
    distances: torch.Tensor,
    mask_gt: torch.Tensor,
    mask_to_ignore: Optional[torch.Tensor] = None,
    cmc_top_k: Tuple[int, ...] = (5,),
    precision_top_k: Tuple[int, ...] = (5,),
    map_top_k: Tuple[int, ...] = (5,),
    reduce: bool = True,
    check_dataset_validity: bool = False,
) -> TMetricsDict:
    """
    Function to count different retrieval metrics.

    Args:
        distances: Distance matrix with the shape of ``[query_size, gallery_size]``
        mask_gt: ``(i,j)`` element indicates if for i-th query j-th gallery is the correct prediction
        mask_to_ignore: Binary matrix to indicate that some of the elements in gallery cannot be used
                     as answers and must be ignored
        cmc_top_k: Tuple of ``k`` values to calculate ``cmc@k`` (`Cumulative Matching Characteristic`)
        precision_top_k: Tuple of  ``k`` values to calculate ``precision@k``
        map_top_k: Tuple of ``k`` values to calculate ``map@k`` (`Mean Average Precision`)
        reduce: If ``False`` return metrics for each query without averaging
        check_dataset_validity: Set ``True`` if you want to check that we have available answers in the gallery for each of the queries

    Returns:
        Metrics dictionary.

    """
    top_k_args = [cmc_top_k, precision_top_k, map_top_k]

    if not any(top_k_args):
        raise ValueError("You must specify arguments for at leas 1 metric to calculate it")

    if check_dataset_validity:
        validate_dataset(mask_gt=mask_gt, mask_to_ignore=mask_to_ignore)

    for top_k_arg in top_k_args:
        if top_k_arg:
            assert all([isinstance(x, int) and (x > 0) for x in top_k_arg])

    if distances.shape != mask_gt.shape:
        raise ValueError(
            f"Distances matrix has the shape of {distances.shape}, "
            f"but mask_to_ignore has the shape of {mask_gt.shape}."
        )

    if (mask_to_ignore is not None) and (mask_to_ignore.shape != distances.shape):
        raise ValueError(
            f"Distances matrix has the shape of {distances.shape}, "
            f"but mask_to_ignore has the shape of {mask_to_ignore.shape}."
        )

    query_sz, gallery_sz = distances.shape

    for top_k_arg in top_k_args:
        for k in top_k_arg:
            if k > gallery_sz:
                warnings.warn(
                    f"Your desired k={k} more than gallery_size={gallery_sz}."
                    f"We'll calculate metrics with k limited by the gallery size."
                )

    if mask_to_ignore is not None:
        distances, mask_gt = apply_mask_to_ignore(distances=distances, mask_gt=mask_gt, mask_to_ignore=mask_to_ignore)

    cmc_top_k_clipped = clip_max(cmc_top_k, gallery_sz)
    precision_top_k_clipped = clip_max(precision_top_k, gallery_sz)
    map_top_k_clipped = clip_max(map_top_k, gallery_sz)

    max_k = max([*cmc_top_k, *precision_top_k, *map_top_k])
    max_k = min(max_k, gallery_sz)

    _, ii_top_k = torch.topk(distances, k=max_k, largest=False)
    ii_arange = torch.arange(query_sz).unsqueeze(-1).expand(query_sz, max_k)
    gt_tops = mask_gt[ii_arange, ii_top_k]

    metrics: TMetricsDict = defaultdict(dict)

    for k_cmc, k_cmc_show in zip(cmc_top_k_clipped, cmc_top_k):
        cmc = torch.any(gt_tops[:, :k_cmc], dim=1).float()
        metrics["cmc"][k_cmc_show] = cmc

    for k_precision, k_precision_show in zip(precision_top_k_clipped, precision_top_k):
        n_gt_matrix = torch.min(mask_gt.sum(dim=1), torch.tensor(k_precision).unsqueeze(0))
        precision = torch.sum(gt_tops[:, :k_precision].float(), dim=1) / n_gt_matrix
        metrics["precision"][k_precision_show] = precision

    for k_map, k_map_show in zip(map_top_k_clipped, map_top_k):
        n_gt_matrix = torch.min(mask_gt.sum(dim=1), torch.tensor(k_map).unsqueeze(0))
        correct_preds = torch.cumsum(gt_tops[:, :k_map], dim=1)
        positions = torch.arange(1, k_map + 1).unsqueeze(0)
        mean_ap = torch.sum((correct_preds / positions) * gt_tops[:, :k_map], dim=1) / n_gt_matrix
        metrics["map"][k_map_show] = mean_ap

    if reduce:
        metrics = reduce_metrics(metrics)

    return metrics


def reduce_metrics(metrics_to_reduce: TMetricsDict) -> TMetricsDict:
    output: TMetricsDict = {}

    for k, v in metrics_to_reduce.items():
        if isinstance(v, (torch.Tensor, np.ndarray)):
            output[k] = v.mean()
        elif isinstance(v, (float, int)):
            output[k] = v
        else:
            output[k] = reduce_metrics(v)  # type: ignore

    return output


def apply_mask_to_ignore(
    distances: torch.Tensor, mask_gt: torch.Tensor, mask_to_ignore: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    distances[mask_to_ignore] = float("inf")
    mask_gt[mask_to_ignore] = False
    return distances, mask_gt


def calc_gt_mask(
    labels: Union[np.ndarray, torch.Tensor],
    is_query: Union[np.ndarray, torch.Tensor],
    is_gallery: Union[np.ndarray, torch.Tensor],
) -> torch.Tensor:
    assert all(isinstance(vector, (np.ndarray, torch.Tensor)) for vector in [labels, is_query, is_gallery])
    assert labels.ndim == is_query.ndim == is_gallery.ndim == 1
    assert len(labels) == len(is_query) == len(is_gallery)

    labels, is_query, is_gallery = map(_to_tensor, [labels, is_query, is_gallery])

    query_mask = is_query == 1
    gallery_mask = is_gallery == 1
    query_labels = labels[query_mask]
    gallery_labels = labels[gallery_mask]
    gt_mask = query_labels[..., None] == gallery_labels[None, ...]

    return gt_mask


def calc_mask_to_ignore(
    is_query: Union[np.ndarray, torch.Tensor], is_gallery: Union[np.ndarray, torch.Tensor]
) -> torch.Tensor:
    assert all(isinstance(vector, (np.ndarray, torch.Tensor)) for vector in [is_query, is_gallery])
    assert is_query.ndim == is_gallery.ndim == 1
    assert len(is_query) == len(is_gallery)

    is_query, is_gallery = map(_to_tensor, [is_query, is_gallery])

    ids_query = torch.nonzero(is_query).squeeze()
    ids_gallery = torch.nonzero(is_gallery).squeeze()
    mask_to_ignore = ids_query[..., None] == ids_gallery[None, ...]

    return mask_to_ignore


def calc_distance_matrix(
    embeddings: Union[np.ndarray, torch.Tensor],
    is_query: Union[np.ndarray, torch.Tensor],
    is_gallery: Union[np.ndarray, torch.Tensor],
) -> torch.Tensor:
    assert all(isinstance(vector, (np.ndarray, torch.Tensor)) for vector in [embeddings, is_query, is_gallery])
    assert is_query.ndim == 1 and is_gallery.ndim == 1 and embeddings.ndim == 2
    assert embeddings.shape[0] == len(is_query) == len(is_gallery)

    embeddings, is_query, is_gallery = map(_to_tensor, [embeddings, is_query, is_gallery])

    query_mask = is_query == 1
    gallery_mask = is_gallery == 1
    query_embeddings = embeddings[query_mask]
    gallery_embeddings = embeddings[gallery_mask]

    distance_matrix = pairwise_dist(x1=query_embeddings, x2=gallery_embeddings, p=2)

    return distance_matrix


def calculate_accuracy_on_triplets(embeddings: torch.Tensor, reduce_mean: bool = True) -> torch.Tensor:
    assert embeddings.ndim == 2
    assert embeddings.shape[0] % 3 == 0

    anchor_ii, positive_ii, negative_ii = get_tri_ids_in_plain(n=len(embeddings))

    pos_dists = elementwise_dist(x1=embeddings[anchor_ii], x2=embeddings[positive_ii]).squeeze()
    neg_dists = elementwise_dist(x1=embeddings[anchor_ii], x2=embeddings[negative_ii]).squeeze()

    acc = (pos_dists < neg_dists).float()

    if reduce_mean:
        return acc.mean()
    else:
        return acc


def validate_dataset(mask_gt: torch.Tensor, mask_to_ignore: torch.Tensor) -> None:
    assert (mask_gt & ~mask_to_ignore).any(1).all(), "There are queries without available correct answers in the gallery!"  # type: ignore


def _to_tensor(array: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
    if isinstance(array, torch.Tensor):
        return array
    elif isinstance(array, np.ndarray):
        return torch.from_numpy(array)
    else:
        raise TypeError("Wrong type")


__all__ = [
    "TMetricsDict",
    "calc_retrieval_metrics",
    "apply_mask_to_ignore",
    "calc_gt_mask",
    "calc_mask_to_ignore",
    "calc_distance_matrix",
    "calculate_accuracy_on_triplets",
    "reduce_metrics",
]



import torch
from torch import nn

from oml.utils.misc import one_hot


class IdealOneHotModel(nn.Module):
    def __init__(self, emb_dim: int, shift: int = 0):
        super(IdealOneHotModel, self).__init__()
        self.emb_dim = emb_dim
        self.shift = shift

    def forward(self, labels: torch.Tensor) -> torch.Tensor:
        embeddings = torch.stack([one_hot(label + self.shift, self.emb_dim) for label in labels])
        return embeddings


class IdealClusterEncoder(nn.Module):
    def forward(self, labels: torch.Tensor, need_noise: bool = True) -> torch.Tensor:
        embeddings = labels + need_noise * 0.01 * torch.randn_like(labels, dtype=torch.float)
        embeddings = embeddings.view((len(labels), 1))
        return embeddings



import hydra
from omegaconf import DictConfig

from oml.const import MOCK_DATASET_PATH
from oml.lightning.entrypoints.validate import pl_val
from oml.utils.download_mock_dataset import download_mock_dataset
from oml.utils.misc import dictconfig_to_dict


@hydra.main(config_path="configs", config_name="val_mock.yaml")
def main_hydra(cfg: DictConfig) -> None:
    cfg = dictconfig_to_dict(cfg)
    download_mock_dataset(MOCK_DATASET_PATH)
    cfg["dataset_root"] = MOCK_DATASET_PATH
    pl_val(cfg)


if __name__ == "__main__":
    main_hydra()



import numpy as np
import torch

from oml.utils.misc_torch import elementwise_dist


def test_elementwise_dist() -> None:
    x1 = torch.randn(size=(3, 4))
    x2 = torch.randn(size=(3, 4))

    val_torch = elementwise_dist(x1=x1, x2=x2, p=2)
    val_custom = np.sqrt(((np.array(x1) - np.array(x2)) ** 2).sum(axis=1))

    assert torch.isclose(val_torch, torch.tensor(val_custom)).all()



from random import randint
from typing import List, Tuple

import pytest
import torch

from oml.interfaces.miners import TLabels
from oml.miners.inbatch_hard_cluster import HardClusterMiner
from oml.miners.inbatch_hard_tri import HardTripletsMiner


@pytest.mark.parametrize(
    ["labels", "expected"],
    [
        [
            [0, 0, 1, 2, 2, 1],
            torch.tensor(
                [
                    [True, True, False, False, False, False],
                    [False, False, True, False, False, True],
                    [False, False, False, True, True, False],
                ]
            ),
        ],
        [
            [1, 2, 3],
            torch.tensor([[True, False, False], [False, True, False], [False, False, True]]),
        ],
        [
            [1, 1, 1, 1, 2, 2, 2, 2],
            torch.tensor(
                [
                    [True, True, True, True, False, False, False, False],
                    [False, False, False, False, True, True, True, True],
                ]
            ),
        ],
    ],
)
def test_cluster_get_labels_mask(labels: List[int], expected: torch.Tensor) -> None:
    """
    Test _get_labels_mask method of HardClusterMiner.

    Args:
        labels: List of labels -- input data for method _skip_diagonal
        expected: Correct answer for labels input

    """
    miner = HardClusterMiner()
    labels_mask = miner._get_labels_mask(labels)
    assert (labels_mask == expected).all()


@pytest.mark.parametrize(
    ["features", "expected"],
    [
        [
            torch.tensor(
                [
                    [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 3]],
                    [[0, 3, 0, 1], [0, 6, 0, 1], [0, 3, 0, 1]],
                ],
                dtype=torch.float,
            ),
            torch.tensor([[1, 1, 4], [1, 4, 1]]),
        ],
        [
            torch.tensor(
                [[[1, 1, 1], [1, 3, 1]], [[2, 2, 6], [2, 6, 2]], [[3, 3, 3], [3, 3, 9]]],
                dtype=torch.float,
            ),
            torch.tensor([[[1, 1], [8, 8], [9, 9]]]),
        ],
    ],
)
def test_cluster_count_intra_label_distances(features: torch.Tensor, expected: torch.Tensor) -> None:
    """
    Test _count_intra_label_distances method of HardClusterMiner.

    Args:
        features: Tensor of shape (n_labels, n_instances, embed_dim)
        embed_dim is an embedding size -- features grouped by labels
        expected: Tensor of shape (n_labels, n_instances) -- expected distances from mean
        vectors of labels to corresponding features
    """
    miner = HardClusterMiner()
    mean_vectors = features.mean(1)
    distances = miner._count_intra_label_distances(features, mean_vectors)
    assert (distances == expected).all()


@pytest.mark.parametrize(
    ["mean_vectors", "expected"],
    [
        [
            torch.tensor([[1, 0, 0, 0], [1, 1, 0, 0], [0, 1, 1, 0]], dtype=torch.float),
            torch.tensor([[0, 1, 3], [1, 0, 2], [3, 2, 0]], dtype=torch.float) ** 0.5,
        ],
        [
            torch.tensor(
                [[0, 0, 0, 0], [3, 0, 0, 0], [0, 4, 0, 0], [0, 0, 0, 5]],
                dtype=torch.float,
            ),
            torch.tensor(
                [[0, 9, 16, 25], [9, 0, 25, 34], [16, 25, 0, 41], [25, 34, 41, 0]],
                dtype=torch.float,
            )
            ** 0.5,
        ],
    ],
)
def test_cluster_count_inter_label_distances(mean_vectors, expected) -> None:  # type: ignore
    """
    Test _count_inter_label_distances method of HardClusterMiner.

    Args:
        mean_vectors: Tensor of shape (n_labels, embed_dim) -- mean vectors of
        labels in the batch
        expected: Tensor of shape (n_labels, n_labels) -- expected distances from mean
        vectors of labels

    """
    miner = HardClusterMiner()
    distances = miner._count_inter_label_distances(mean_vectors)
    assert (distances == expected).all()


@pytest.mark.parametrize(
    ["embed_dim", "labels", "expected_shape"],
    [
        [128, [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3], [(4, 128), (4, 128), (4, 128)]],
        [32, [1, 2, 3, 4, 5, 1, 2, 3, 4, 5], [(5, 32), (5, 32), (5, 32)]],
        [16, torch.tensor([0, 0, 1, 1]), [(2, 16), (2, 16), (2, 16)]],
    ],
)
def test_cluster_sample_shapes(embed_dim: int, labels: TLabels, expected_shape: List[Tuple[int]]) -> None:
    """
    Test output shapes in sample method of HardClusterMiner.

    Args:
        embed_dim: Size of embedding
        labels: List of labels for samples in batch
        expected_shape: Expected shape of output triplet

    """
    miner = HardClusterMiner()
    batch_size = len(labels)
    features = torch.rand(size=(batch_size, embed_dim))
    anchor, positive, negative = miner.sample(features, labels)
    anchor_shape, pos_shape, neg_shape = expected_shape

    assert anchor.shape == anchor_shape
    assert positive.shape == pos_shape
    assert negative.shape == neg_shape


def test_triplet_cluster_edge_case() -> None:
    """
    Check an edge case of trivial samples for labels:
    expected HardTripletsMiner and HardClusterMiner to
    generate the same triplets.

    """
    features_dim = 128
    n_labels, n_instances = randint(2, 32), randint(2, 32)

    # Create a list of random labels
    unique_labels = torch.tensor(list(range(n_labels)))
    # Create a list of random features for all the labels
    unique_features = torch.rand(size=(n_labels, features_dim), dtype=torch.float)

    labels = unique_labels.repeat(n_instances)
    features = unique_features.repeat((n_instances, 1))

    hard_triplet_miner = HardTripletsMiner()
    hard_cluster_miner = HardClusterMiner()

    triplets = hard_triplet_miner.sample(features, labels)
    cluster_triplets = hard_cluster_miner.sample(features, labels)

    # Concatenates tensors from triplets to use torch.unique for comparison
    triplets = torch.cat(triplets, dim=1)
    cluster_triplets = torch.cat(cluster_triplets, dim=1)

    triplets = torch.unique(triplets, dim=0)
    cluster_triplets = torch.unique(cluster_triplets, dim=0)

    assert torch.allclose(triplets, cluster_triplets, atol=1e-10)



from collections import Counter
from typing import List

from scipy.special import binom


def check_triplets_consistency(
    ids_anchor: List[int], ids_pos: List[int], ids_neg: List[int], labels: List[int]
) -> None:
    """
    Args:
        ids_anchor: Anchor indexes of selected triplets
        ids_pos: Positive indexes of selected triplets
        ids_neg: Negative indexes of selected triplets
        labels: Labels of the samples in the batch

    """
    num_sampled_tri = len(ids_anchor)

    assert num_sampled_tri == len(ids_pos) == len(ids_neg)

    for (i_a, i_p, i_n) in zip(ids_anchor, ids_pos, ids_neg):
        assert len({i_a, i_p, i_n}) == 3
        assert labels[i_a] == labels[i_p]
        assert labels[i_a] != labels[i_n]

    unq_tri = set(zip(ids_anchor, ids_pos, ids_neg))

    assert num_sampled_tri == len(unq_tri)


def calc_n_triplets(labels: List[int]) -> int:
    labels_counts = Counter(labels).values()

    n_all_tri = 0
    for count in labels_counts:
        n_pos = 2 * binom(count, 2)
        n_neg = len(labels) - count
        n_all_tri += n_pos * n_neg

    return n_all_tri


def check_all_triplets_number(labels: List[int], num_selected_tri: int, max_tri: int) -> None:
    """
    Checks that the selection strategy for all triplets
    returns the correct number of triplets.

    Args:
        labels: List of labels
        num_selected_tri: Number of selected triplets
        max_tri: Limit on the number of selected triplets

    """
    n_all_tri = calc_n_triplets(labels=labels)
    assert num_selected_tri == n_all_tri or num_selected_tri == max_tri



# flake8: noqa
from oml.miners.inbatch_all_tri import AllTripletsMiner
from tests.test_oml.test_miners.shared_checkers import (
    check_all_triplets_number,
    check_triplets_consistency,
)


def test_all_triplets_miner(features_and_labels) -> None:  # type: ignore
    """
    Args:
        features_and_labels: Features and valid labels

    """
    max_tri = 512
    miner = AllTripletsMiner(max_output_triplets=max_tri)

    for _, labels in features_and_labels:
        ids_a, ids_p, ids_n = miner._sample(labels=labels)

        check_all_triplets_number(labels=labels, max_tri=max_tri, num_selected_tri=len(ids_a))

        check_triplets_consistency(ids_anchor=ids_a, ids_pos=ids_p, ids_neg=ids_n, labels=labels)



from typing import Any, Dict

from oml.transforms.images.albumentations.transforms import (
    get_augs_albu,
    get_normalisation_albu,
    get_normalisation_resize_albu,
    get_normalisation_resize_albu_clip,
)
from oml.transforms.images.torchvision.transforms import (
    get_augs_hypvit,
    get_augs_torch,
    get_normalisation_resize_hypvit,
    get_normalisation_resize_torch,
    get_normalisation_torch,
)
from oml.transforms.images.utils import TTransforms
from oml.utils.misc import TCfg, dictconfig_to_dict

TRANSFORMS_ALBU = {
    "augs_albu": get_augs_albu,
    "norm_albu": get_normalisation_albu,
    "norm_resize_albu": get_normalisation_resize_albu,
    "norm_resize_albu_clip": get_normalisation_resize_albu_clip,
}

TRANSFORMS_TORCH = {
    "augs_torch": get_augs_torch,
    "norm_torch": get_normalisation_torch,
    "norm_resize_torch": get_normalisation_resize_torch,
    "augs_hypvit_torch": get_augs_hypvit,
    "norm_resize_hypvit_torch": get_normalisation_resize_hypvit,
}

TRANSFORMS_REGISTRY = {**TRANSFORMS_ALBU, **TRANSFORMS_TORCH}


def get_transforms(name: str, **kwargs: Dict[str, Any]) -> TTransforms:
    augs = TRANSFORMS_REGISTRY[name](**kwargs)  # type: ignore
    return augs


def get_transforms_by_cfg(cfg: TCfg) -> TTransforms:
    cfg = dictconfig_to_dict(cfg)
    return get_transforms(name=cfg["name"], **cfg["args"])


__all__ = [
    "TRANSFORMS_TORCH",
    "TRANSFORMS_ALBU",
    "TRANSFORMS_REGISTRY",
    "get_transforms",
    "get_transforms_by_cfg",
]



import math
from collections import Counter, defaultdict
from functools import partial
from operator import itemgetter
from random import randint, shuffle
from typing import Any, Dict, List, Set, Tuple, Union

import pytest

from oml.samplers.category_balance import CategoryBalanceSampler
from oml.samplers.distinct_category_balance import DistinctCategoryBalanceSampler
from oml.utils.misc import set_global_seed

TLabelsWithMapping = List[Tuple[List[int], Dict[int, int]]]


def generate_valid_categories_labels(num: int, guarantee_enough_labels: bool = True) -> TLabelsWithMapping:
    """
    This function generates some valid inputs for category sampler.

    Args:
        num: Number of samples to generate
        guarantee_enough_labels: If you want to guarantee enough amount of labels

    Returns:
        Labels with the mapping from labels to categories

    """
    labels_with_mapping: TLabelsWithMapping = []

    for _ in range(num):
        unique_labels_number = randint(35, 55)
        n_labels, n_instances = randint(3, 7), randint(5, 10)
        unique_labels = list(range(unique_labels_number))
        shuffle(unique_labels)
        label2category = {}
        idx = 0
        cat = 0
        while idx < unique_labels_number:
            if guarantee_enough_labels:
                new_idx = idx + randint(0, 5) + n_labels
                labels_subset = unique_labels[idx:new_idx]
                if len(labels_subset) < n_labels:
                    cat -= 1
            else:
                new_idx = idx + randint(2, n_labels - 1)
                labels_subset = unique_labels[idx:new_idx]
                # process last segment of labels
                if len(labels_subset) == 1:
                    cat -= 1
            label2category.update({label: cat for label in labels_subset})
            idx = new_idx
            cat += 1
        labels = []
        for label in unique_labels:
            labels.extend([label] * randint(2, 20))
        shuffle(labels)
        labels_with_mapping.append((labels, label2category))
    return labels_with_mapping


def get_valid_batch_params(labels: List[int], label2category: Dict[int, int]) -> Tuple[int, int, int]:
    # labels
    category2label = defaultdict(list)
    for label, category in label2category.items():
        category2label[category].append(label)
    categories_sizes = list(map(len, category2label.values()))
    n_labels = randint(2, min(categories_sizes))

    # categories
    n_categories = randint(1, len(set(label2category.values())))

    # instances
    n_instances = randint(2, max(Counter(labels).values()))

    return n_labels, n_instances, n_categories


@pytest.fixture()
def input_for_category_balance_batch_sampler_few_labels() -> TLabelsWithMapping:
    # (julia-shenshina) It was checked once with N = 100_000 before doing the PR
    num_random_cases = 100
    input_cases = generate_valid_categories_labels(num_random_cases, guarantee_enough_labels=False)
    return input_cases


@pytest.fixture()
def input_for_category_balance_batch_sampler() -> TLabelsWithMapping:
    # (julia-shenshina) It was checked once with N = 100_000 before doing the PR
    num_random_cases = 100
    input_cases = generate_valid_categories_labels(num_random_cases, guarantee_enough_labels=True)
    return input_cases


def check_category_balance_batch_sampler_epoch(
    sampler: Union[CategoryBalanceSampler, DistinctCategoryBalanceSampler],
    labels: List[int],
    label2category: Dict[int, int],
    resample_labels: bool,
    epoch_size: int,
) -> None:
    sampled_ids = list(sampler)

    collected_labels: Set[int] = set()
    collected_categories: Set[int] = set()

    # emulating of 1 epoch
    for i, batch_ids in enumerate(sampled_ids):
        assert len(batch_ids) == sampler.n_labels * sampler.n_categories * sampler.n_instances

        batch_labels = itemgetter(*batch_ids)(labels)  # type: ignore
        batch_categories = set(label2category[label] for label in batch_labels)

        # check that we sampled n_categories at all the batches
        assert len(batch_categories) == sampler.n_categories

        # check that new sampled at most n_labels at all the batches
        if resample_labels:
            assert len(set(batch_labels)) <= sampler.n_labels * sampler.n_categories
        else:
            assert len(set(batch_labels)) == sampler.n_labels * sampler.n_categories

        collected_labels.update(batch_labels)
        collected_categories.update(batch_categories)

        labels_counter = Counter(batch_labels)
        num_batch_labels = len(labels_counter)
        num_batch_samples = list(labels_counter.values())
        cur_batch_size = len(batch_labels)

        # batch-level invariants
        assert len(set(batch_ids)) >= 4, set(batch_ids)  # type: ignore

        assert num_batch_labels <= sampler.n_categories * sampler.n_labels, (
            num_batch_labels,
            sampler.n_categories,
            sampler.n_labels,
        )

        if resample_labels:
            assert all(el >= sampler.n_instances for el in num_batch_samples)
        else:
            assert all(el == sampler.n_instances for el in num_batch_samples)

        assert cur_batch_size == sampler.n_categories * sampler.n_labels * sampler.n_instances

    # epoch-level invariants
    assert len(collected_categories) <= len(set(label2category.values()))
    assert len(sampler) == epoch_size


@pytest.mark.parametrize(
    "sampler_class,sampler_kwargs",
    (
        (CategoryBalanceSampler, {"resample_labels": False}),
        (DistinctCategoryBalanceSampler, {"epoch_size": 20}),
    ),
)
def test_category_batch_sampler_resample_raises(sampler_class: Any, sampler_kwargs: Dict[str, Any]) -> None:
    """
    Check the behavior of samplers in case of lack of labels in a category.

    """
    label2category = {0: 0, 1: 0, 2: 0, 3: 0, 4: 1, 5: 1}
    labels = [0] * 5 + [1] * 4 + [2] * 6 + [3] * 5 + [4] * 5 + [5] * 6
    n_categories, n_labels, n_instances = 2, 3, 2
    with pytest.raises(ValueError, match="All the categories must have at least 3 unique labels"):
        _ = sampler_class(
            labels=labels,
            label2category=label2category,
            n_categories=n_categories,
            n_labels=n_labels,
            n_instances=n_instances,
            **sampler_kwargs
        )

    assert True


@pytest.mark.parametrize(
    "fixture_name,resample_labels",
    (
        # valid input: enough categories and labels for sampling without repetition
        ("input_for_category_balance_batch_sampler", False),
        ("input_for_category_balance_batch_sampler", True),
        # not enough labels in some categories
        ("input_for_category_balance_batch_sampler_few_labels", True),
    ),
)
def test_category_balance_batch_sampler(
    request: pytest.FixtureRequest, fixture_name: str, resample_labels: bool
) -> None:
    for labels, label2category in request.getfixturevalue(fixture_name):
        n_labels, n_instances, n_categories = get_valid_batch_params(labels, label2category)

        sampler = CategoryBalanceSampler(
            labels=labels,
            label2category=label2category,
            resample_labels=resample_labels,
            n_categories=n_categories,
            n_labels=n_labels,
            n_instances=n_instances,
        )
        check_category_balance_batch_sampler_epoch(
            sampler=sampler,
            labels=labels,
            label2category=label2category,
            resample_labels=resample_labels,
            epoch_size=math.ceil(len(set(labels)) / n_labels),
        )

    assert True


def test_category_balance_batch_sampler_policy(input_for_category_balance_batch_sampler: TLabelsWithMapping) -> None:
    """
    Check that sampling with categories behaves the same in case
    of valid input data with enough labels for all the categories for resample_labels "resample" and "raise".

    """
    for labels, label2category in input_for_category_balance_batch_sampler:
        n_labels, n_instances, n_categories = get_valid_batch_params(labels, label2category)

        sampler = CategoryBalanceSampler(
            labels=labels,
            label2category=label2category,  # type: ignore
            n_labels=n_labels,
            n_instances=n_instances,
            n_categories=n_categories,
            resample_labels=True,
        )
        check_category_balance_batch_sampler_epoch(
            sampler=sampler,
            labels=labels,
            label2category=label2category,
            resample_labels=False,
            epoch_size=math.ceil(len(set(labels)) / n_labels),
        )

    assert True


@pytest.mark.parametrize(
    "fixture_name,epoch_size",
    (
        # valid input: enough categories and labels for sampling without repetition
        ("input_for_category_balance_batch_sampler", 100),
    ),
)
def test_distinct_category_balance_batch_sampler(
    request: pytest.FixtureRequest, fixture_name: str, epoch_size: int
) -> None:
    fixture = request.getfixturevalue(fixture_name)
    for labels, label2category in fixture:
        n_labels, n_instances, n_categories = get_valid_batch_params(labels, label2category)

        sampler = DistinctCategoryBalanceSampler(
            labels=labels,
            label2category=label2category,
            epoch_size=epoch_size,
            n_labels=n_labels,
            n_instances=n_instances,
            n_categories=n_categories,
        )
        check_category_balance_batch_sampler_epoch(
            sampler=sampler,
            labels=labels,
            label2category=label2category,
            resample_labels=True,
            epoch_size=epoch_size,
        )

    assert True


@pytest.mark.parametrize(
    "sampler_constructor",
    [CategoryBalanceSampler, partial(DistinctCategoryBalanceSampler, epoch_size=50)],
)
def test_categories_as_strings(sampler_constructor) -> None:  # type: ignore
    labels, label2category = generate_valid_categories_labels(1, guarantee_enough_labels=True)[0]

    n_labels, n_instances, n_categories = get_valid_batch_params(labels, label2category)

    label2category_str = {label: str(cat) for label, cat in label2category.items()}

    set_global_seed(0)
    sampler = sampler_constructor(labels, label2category, n_categories, n_labels, n_instances)
    ii_sampled = list(iter(sampler))

    set_global_seed(0)
    sampler_with_str = sampler_constructor(labels, label2category_str, n_categories, n_labels, n_instances)
    ii_sampled_with_str = list(iter(sampler_with_str))

    assert ii_sampled == ii_sampled_with_str, list(zip(ii_sampled, ii_sampled_with_str))



from typing import Optional

import pandas as pd
import pytest
from omegaconf import OmegaConf

from oml.const import CONFIGS_PATH, MOCK_DATASET_PATH
from oml.datasets.base import DatasetWithLabels
from oml.registry.transforms import TRANSFORMS_REGISTRY, get_transforms_by_cfg
from oml.transforms.images.utils import get_im_reader_for_transforms


@pytest.mark.parametrize("aug_name", list(TRANSFORMS_REGISTRY.keys()))
def test_transforms(aug_name: Optional[str]) -> None:
    df = pd.read_csv(MOCK_DATASET_PATH / "df.csv")
    transforms = get_transforms_by_cfg(OmegaConf.load(CONFIGS_PATH / "transforms" / f"{aug_name}.yaml"))
    f_imread = get_im_reader_for_transforms(transforms)

    dataset = DatasetWithLabels(df=df, dataset_root=MOCK_DATASET_PATH, transform=transforms, f_imread=f_imread)

    _ = dataset[0]

    assert True


def test_default_transforms() -> None:
    df = pd.read_csv(MOCK_DATASET_PATH / "df.csv")
    dataset = DatasetWithLabels(df=df, dataset_root=MOCK_DATASET_PATH, transform=None)
    _ = dataset[0]
    assert True



# type: ignore
from itertools import chain
from math import ceil
from random import randint
from typing import Any, Dict, List

import pytest
from torch.utils.data import DataLoader

from oml.ddp.patching import patch_dataloader_to_ddp
from oml.ddp.utils import sync_dicts_ddp
from oml.interfaces.samplers import IBatchSampler
from oml.samplers.balance import BalanceSampler
from oml.samplers.category_balance import CategoryBalanceSampler
from oml.samplers.distinct_category_balance import DistinctCategoryBalanceSampler

from .utils import init_ddp, run_in_ddp


@pytest.mark.parametrize("n_labels_sampler", [2, 5])
@pytest.mark.parametrize("n_instances_sampler", [2, 5])
@pytest.mark.parametrize("n_labels_dataset", [100, 85])
@pytest.mark.parametrize(
    "sampler_class, setup_kwargs",
    [
        (BalanceSampler, {}),
        (CategoryBalanceSampler, {"n_categories": 2, "num_categories_in_dataset": 10}),
        (CategoryBalanceSampler, {"n_categories": 3, "num_categories_in_dataset": 17}),
        (DistinctCategoryBalanceSampler, {"n_categories": 2, "num_categories_in_dataset": 10, "epoch_size": 4}),
        (DistinctCategoryBalanceSampler, {"n_categories": 4, "num_categories_in_dataset": 17, "epoch_size": 5}),
    ],
)
@pytest.mark.parametrize("world_size", [1, 2, 3])
def test_patching_balance_sampler(
    world_size: int,
    n_labels_dataset: int,
    n_labels_sampler: int,
    n_instances_sampler: int,
    sampler_class: IBatchSampler,
    setup_kwargs: Dict[str, Any],
) -> None:
    args = (n_labels_dataset, n_labels_sampler, n_instances_sampler, sampler_class, setup_kwargs)
    run_in_ddp(world_size=world_size, fn=check_patching_balance_batch_sampler, args=args)


def setup_batch_sampler(
    sampler_class: IBatchSampler, labels: List[int], n_labels: int, n_instances: int, **kwargs: Dict[str, Any]
) -> IBatchSampler:
    class2setup = {
        BalanceSampler: _setup_balance_sampler,
        CategoryBalanceSampler: _setup_category_sampler,
        DistinctCategoryBalanceSampler: _setup_distinct_category_sampler,
    }

    return class2setup[sampler_class](labels=labels, n_labels=n_labels, n_instances=n_instances, **kwargs)


def _setup_balance_sampler(
    labels: List[int], n_labels: int, n_instances: int, **kwargs: Dict[str, Any]
) -> BalanceSampler:
    return BalanceSampler(labels=labels, n_labels=n_labels, n_instances=n_instances)


def _setup_category_sampler(
    labels: List[int],
    n_labels: int,
    n_instances: int,
    n_categories: int,
    num_categories_in_dataset: int,
    **kwargs: Dict[str, Any]
) -> CategoryBalanceSampler:
    label2category = dict(zip(labels, [label % num_categories_in_dataset for label in labels]))
    return CategoryBalanceSampler(
        labels=labels,
        label2category=label2category,
        n_categories=n_categories,
        n_labels=n_labels,
        n_instances=n_instances,
    )


def _setup_distinct_category_sampler(
    labels: List[int],
    n_labels: int,
    n_instances: int,
    n_categories: int,
    num_categories_in_dataset: int,
    epoch_size: int,
    **kwargs: Dict[str, Any]
) -> DistinctCategoryBalanceSampler:
    label2category = dict(zip(labels, [label % num_categories_in_dataset for label in labels]))
    return DistinctCategoryBalanceSampler(
        labels=labels,
        label2category=label2category,
        n_categories=n_categories,
        n_labels=n_labels,
        n_instances=n_instances,
        epoch_size=epoch_size,
    )


def check_patching_balance_batch_sampler(
    rank: int,
    world_size: int,
    n_labels_dataset: int,
    n_labels_sampler: int,
    n_instances_sampler: int,
    sampler_class,
    setup_kwargs,
) -> None:
    init_ddp(rank, world_size)
    labels = [[label] * randint(n_instances_sampler, 2 * n_instances_sampler) for label in range(n_labels_dataset)]
    labels = list(chain(*labels))
    dataset = list(range(len(labels)))

    batch_sampler = setup_batch_sampler(
        sampler_class, labels=labels, n_labels=n_labels_sampler, n_instances=n_instances_sampler, **setup_kwargs
    )

    loader = DataLoader(dataset=dataset, batch_sampler=batch_sampler, collate_fn=lambda x: tuple(x))

    loader_ddp = patch_dataloader_to_ddp(loader)

    outputs_from_epochs = []

    for epoch in range(3):
        outputs = []
        for batch in loader_ddp:
            outputs.append(batch)

        outputs_from_epochs.append(outputs)

        seq_outputs = list(chain(*outputs))

        # number of batches should be divided by the number of devices with padding
        expected_len_per_process = ceil(len(loader) / max(1, world_size))
        assert len(outputs) == expected_len_per_process

        outputs_synced = sync_dicts_ddp({"batches": outputs}, world_size)["batches"]

        assert len(outputs_synced) == max(1, world_size) * expected_len_per_process

        if sampler_class == BalanceSampler:
            # Check each batches without repeating of ids
            assert len(set(seq_outputs)) == len(seq_outputs)

            _num_batches_after_ddp_padding = expected_len_per_process * max(1, world_size)
            possible_num_not_unique_batches = _num_batches_after_ddp_padding - len(loader)

            assert possible_num_not_unique_batches < world_size
            assert len(outputs_synced) - len(set(outputs_synced)) <= possible_num_not_unique_batches

    # we check that batches on each epoch are different
    outputs_from_epochs = list(map(tuple, outputs_from_epochs))
    assert len(set(outputs_from_epochs)) == len(outputs_from_epochs)


@pytest.mark.parametrize("shuffle", [True, False])
@pytest.mark.parametrize("drop_last", [True, False])
@pytest.mark.parametrize("batch_size", [10, 17])
@pytest.mark.parametrize("num_samples", [100, 85])
@pytest.mark.parametrize("world_size", [1, 2, 3])
def test_patching_seq_sampler(
    world_size: int, num_samples: int, drop_last: bool, shuffle: bool, batch_size: int, num_workers: int
) -> None:
    args = (num_samples, drop_last, shuffle, batch_size, num_workers)
    run_in_ddp(world_size=world_size, fn=check_patching_seq_sampler, args=args)


def check_patching_seq_sampler(
    rank: int, world_size: int, num_samples: int, drop_last: bool, shuffle: bool, batch_size: int, num_workers: int
) -> None:
    init_ddp(rank, world_size)
    dataset = list(range(num_samples))
    loader = DataLoader(
        dataset=dataset,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        batch_size=batch_size,
        collate_fn=lambda x: x,
    )

    loader_ddp = patch_dataloader_to_ddp(loader)

    outputs_from_epochs = []

    for epoch in range(2):
        outputs = []
        for batch in loader_ddp:
            outputs.extend(batch)

        outputs_from_epochs.append(outputs)

        if world_size > 1:
            expected_len_per_process = ceil(num_samples / world_size)
            if drop_last:
                expected_len_per_process = expected_len_per_process // batch_size * batch_size
        else:
            expected_len_per_process = num_samples // batch_size * batch_size if drop_last else num_samples

        # check that ids are uniques and amount of them is reduced
        assert len(outputs) == len(set(outputs)) == expected_len_per_process

        outputs_synced = sync_dicts_ddp({"ids": outputs}, world_size)["ids"]

        assert len(outputs_synced) == max(1, world_size) * expected_len_per_process

        _num_samples_after_ddp_padding = ceil(num_samples / max(1, world_size)) * max(1, world_size)
        possible_num_not_unique_samples = _num_samples_after_ddp_padding - num_samples

        assert possible_num_not_unique_samples < world_size
        assert len(outputs_synced) - len(set(outputs_synced)) <= possible_num_not_unique_samples

    # Depends on the shuffle bathes should be the same or different
    outputs_from_epochs = list(map(tuple, outputs_from_epochs))
    if shuffle:
        assert len(set(outputs_from_epochs)) == len(outputs_from_epochs)
    else:
        assert len(set(outputs_from_epochs)) != len(outputs_from_epochs)



from abc import ABC
from collections.abc import MutableMapping
from typing import Any, Dict, Hashable, Iterator, List, Optional, Tuple, Type, Union

import numpy as np
import torch
from torch import Tensor, cdist

TSingleValues = Union[int, float, np.float_, np.int_, torch.Tensor]
TSequenceValues = Union[List[float], Tuple[float, ...], np.ndarray, torch.Tensor]
TOnlineValues = Union[TSingleValues, TSequenceValues]


def elementwise_dist(x1: Tensor, x2: Tensor, p: int = 2) -> Tensor:
    """
    Args:
        x1: tensor with the shape of [N, D]
        x2: tensor with the shape of [N, D]
        p: degree

    Returns: elementwise distances with the shape of [N]

    """
    assert len(x1.shape) == len(x2.shape) == 2
    assert x1.shape == x2.shape

    # we need an extra dim here to avoid pairwise behaviour of torch.cdist
    if len(x1.shape) == 2:
        x1 = x1.unsqueeze(1)
        x2 = x2.unsqueeze(1)

    dist = cdist(x1=x1, x2=x2, p=p).squeeze()

    return dist


def pairwise_dist(x1: Tensor, x2: Tensor, p: int = 2) -> Tensor:
    """
    Args:
        x1: tensor with the shape of [N, D]
        x2: tensor with the shape of [M, D]
        p: degree

    Returns: pairwise distances with the shape of [N, M]

    """
    assert len(x1.shape) == len(x2.shape) == 2
    assert x1.shape[-1] == x2.shape[-1]

    return cdist(x1=x1, x2=x2, p=p)


def _check_is_sequence(val: Any) -> bool:
    try:
        len(val)
        return True
    except Exception:
        return False


class OnlineCalc(ABC):
    """
    The base class to calculate some statistics online (on the steam of values).

    """

    def __init__(self, val: Optional[TOnlineValues] = None):
        self.result: float = 0.0
        if val is not None:
            self.update(val)

    def update(self, val: TOnlineValues) -> None:
        if _check_is_sequence(val):
            self.calc_with_batch(val)
        else:
            self.calc_with_single_value(val)

    def calc_with_single_value(self, val: TSingleValues) -> None:
        """
        Calculation with non iterable types: float, int / numpy and torch elements (array and elements of
        array have different types and methods)

        """
        self.calc_with_batch([val])

    def calc_with_batch(self, val: TSequenceValues) -> None:
        """
        Vectorized calculation of online value

        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.__dict__)})"


class AvgOnline(OnlineCalc):
    def __init__(self, *args: Any, **kwargs: Any):
        self.n = 0
        super().__init__(*args, **kwargs)

    def calc_with_batch(self, val: TSequenceValues) -> None:
        len_val = len(val)
        self.n += len_val
        self.result = sum(val) / self.n + (self.n - len_val) / self.n * self.result  # type: ignore


class SumOnline(OnlineCalc):
    def calc_with_batch(self, val: TSequenceValues) -> None:
        self.result += float(sum(val))


class OnlineDict(MutableMapping):  # type: ignore
    """
    We don't inherite from built-in 'dict' due to internal C optimization. We mimic to dict with MutableMapping
    https://treyhunner.com/2019/04/why-you-shouldnt-inherit-from-list-and-dict-in-python/

    """

    online_calculator: Type[OnlineCalc]

    def __init__(self, input_dict: Optional[Dict[Hashable, TOnlineValues]] = None):
        self.dict: Dict[Hashable, OnlineCalc] = {}

        if input_dict:
            self.update(input_dict)

    def __setitem__(self, key: Hashable, value: TOnlineValues) -> None:
        self.dict[key] = self.online_calculator(value)

    def __getitem__(self, key: Hashable) -> float:
        return self.dict[key].result

    def __len__(self) -> int:
        return len(self.dict)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.dict)

    def __delitem__(self, key: Hashable) -> None:
        return self.dict.__delitem__(key)

    def update(self, input_dict: Dict[Hashable, TOnlineValues]) -> None:  # type: ignore
        for k, v in input_dict.items():
            if k in self.dict:
                self.dict[k].update(v)
            else:
                self[k] = v

    def __repr__(self) -> str:
        output = {k: v.result for k, v in self.dict.items()}
        return f"{self.__class__.__name__}({output})"

    def get_dict_with_results(self) -> Dict[str, float]:
        return {k: v for k, v in self.items()}


class OnlineAvgDict(OnlineDict):
    online_calculator = AvgOnline


class OnlineSumDict(OnlineDict):
    online_calculator = SumOnline


__all__ = [
    "elementwise_dist",
    "pairwise_dist",
    "OnlineCalc",
    "AvgOnline",
    "SumOnline",
    "OnlineDict",
    "OnlineAvgDict",
    "OnlineSumDict",
]



from pathlib import Path
from typing import Any, Dict

import pytest
import torch

from oml.interfaces.models import IExtractor
from oml.models.resnet import ResnetExtractor
from oml.models.vit.clip import ViTCLIPExtractor
from oml.models.vit.vit import ViTExtractor

vit_args = {"normalise_features": False, "use_multi_scale": False, "strict_load": True}
vit_clip_args = {"normalise_features": False, "strict_load": True}
resnet_args = {"normalise_features": True, "gem_p": 7.0, "remove_fc": False, "strict_load": True}


@pytest.mark.parametrize(
    "constructor,args,default_arch,download_large_checkpoints",
    [
        (ViTExtractor, vit_args, "vits16", False),
        (ViTCLIPExtractor, vit_clip_args, "vitb32_224", False),
        (ResnetExtractor, resnet_args, "resnet50", False),
    ],
)
def test_creation(
    constructor: IExtractor, args: Dict[str, Any], default_arch: str, download_large_checkpoints: bool
) -> None:
    # 1. Random weights
    net = constructor(weights=None, arch=default_arch, **args)
    net.forward(torch.randn(1, 3, 224, 224))

    # 2. Load from file
    fname = f"{default_arch}_random.pth"
    torch.save({"state_dict": net.state_dict()}, fname)
    net = constructor(weights=fname, arch=default_arch, **args)
    net(torch.randn(1, 3, 224, 224))
    Path(fname).unlink()

    # 3. Pretrained checkpoints
    for key in constructor.pretrained_models.keys():
        if constructor == ViTCLIPExtractor:
            is_large_model = "vitl" in key
            if is_large_model and not download_large_checkpoints:
                continue
            arch = key.split("_", maxsplit=1)[1]  # sber_vitb16_224 -> vitb16_224, openai_vitb16_224 -> vitb16_224
        else:
            arch = key.split("_")[0]

        net = constructor(weights=key, arch=arch, **args)
        net(torch.randn(1, 3, 224, 224))

    assert True



from argparse import ArgumentParser, Namespace
from itertools import chain
from typing import Any, Dict, List, Tuple

import pandas as pd
import torch
from pytorch_lightning import Trainer
from pytorch_lightning.utilities.types import EPOCH_OUTPUT
from torch import nn
from torch.optim import Adam
from torch.utils.data import DataLoader

from oml.const import INPUT_TENSORS_KEY, MOCK_DATASET_PATH, TMP_PATH
from oml.datasets.triplet import TriDataset, TTriplet, tri_collate
from oml.ddp.utils import sync_dicts_ddp
from oml.lightning.entrypoints.parser import parse_engine_params_from_config
from oml.lightning.modules.module_ddp import (
    ModuleDDP,
    TTrainDataloaders,
    TValDataloaders,
)
from oml.losses.triplet import TripletLossPlain
from oml.miners.inbatch_all_tri import get_available_triplets
from oml.transforms.images.albumentations.transforms import (
    get_normalisation_resize_albu,
)
from oml.utils.download_mock_dataset import download_mock_dataset
from oml.utils.misc import set_global_seed


def get_triplets_from_retrieval_setup(df: pd.DataFrame) -> Tuple[List[TTriplet], List[TTriplet]]:
    outputs = []

    for split in ["train", "validation"]:
        df_split = df[df["split"] == split].reset_index(drop=True)
        ids_apn = get_available_triplets(list(df_split["label"]))
        triplets = list(zip(*map(lambda x: [df_split["path"][idx] for idx in x], ids_apn)))

        outputs.append(triplets)

    return outputs  # type: ignore


class DummyModule(ModuleDDP):
    save_path_ckpt_pattern = str(TMP_PATH / "ckpt_experiment_{experiment}.pth")
    save_path_train_ids_pattern = str(TMP_PATH / "train_ids_{experiment}_{epoch}.pth")
    save_path_val_ids_pattern = str(TMP_PATH / "val_ids_{experiment}_{epoch}.pth")

    def __init__(self, exp_num: int, loaders_val: TValDataloaders, loaders_train: TTrainDataloaders):
        super().__init__(loaders_val=loaders_val, loaders_train=loaders_train)
        self.exp_num = exp_num
        # We use a simple model, instead of ViT or ResNet, because they have Dropouts
        self.model = nn.Sequential(
            nn.AvgPool2d((32, 32)), nn.Flatten(), *([nn.Sigmoid(), nn.Linear(3, 3, bias=False)] * 3)
        )
        self.criterion = TripletLossPlain(margin=None)

    def validation_step(self, batch: Dict[str, Any], batch_idx: int, *dataset_idx: int) -> Dict[str, Any]:
        embeddings = self.model(batch[INPUT_TENSORS_KEY])
        return {**batch, **{"embeddings": embeddings}}

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> Dict[str, Any]:
        embeddings = self.model(batch[INPUT_TENSORS_KEY])
        loss = self.criterion(embeddings)
        batch["loss"] = loss
        return batch

    def check_and_save_ids(self, outputs: EPOCH_OUTPUT, mode: str) -> None:
        assert mode in ("train", "val")
        world_size = self.trainer.world_size

        ids_batches = [[int(idx_str[:-2]) for idx_str in batch_dict["tri_ids"][::3]] for batch_dict in outputs]
        ids_flatten = list(chain(*ids_batches))

        ids_flatten_synced = sync_dicts_ddp({"ids_flatten": ids_flatten}, world_size)["ids_flatten"]
        assert len(ids_flatten_synced) == len(ids_flatten) * world_size == len(set(ids_flatten_synced))

        ids_per_step = {step: ids for step, ids in enumerate(ids_batches)}
        ids_per_step_synced = sync_dicts_ddp(ids_per_step, world_size)  # type: ignore
        ids_per_step_synced = {step: sorted(synced_ids) for step, synced_ids in ids_per_step_synced.items()}

        pattern = self.save_path_train_ids_pattern if mode == "train" else self.save_path_val_ids_pattern
        torch.save(ids_per_step_synced, pattern.format(experiment=self.exp_num, epoch=self.trainer.current_epoch))

    def training_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        self.check_and_save_ids(outputs, "train")

    def validation_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        self.check_and_save_ids(outputs, "val")

    def on_train_end(self) -> None:
        torch.save(self.model, self.save_path_ckpt_pattern.format(experiment=self.exp_num))

    def configure_optimizers(self) -> Any:
        return Adam(params=self.parameters(), lr=1e-3)


def experiment(args: Namespace) -> None:
    download_mock_dataset(MOCK_DATASET_PATH)

    devices = args.devices
    len_dataset = args.len_dataset
    batch_size = args.batch_size
    exp_num = args.exp_num
    max_epochs = args.max_epochs

    assert (
        len_dataset % (batch_size * devices) == 0
    ), "For this experiment, the parameters must be a multiple of each other"

    set_global_seed(1)

    df = pd.read_csv(MOCK_DATASET_PATH / "df.csv")
    triplets_train, triplets_val = get_triplets_from_retrieval_setup(df)

    assert len(triplets_train) > len_dataset
    assert len(triplets_val) > len_dataset
    triplets_train = triplets_train[:len_dataset]
    triplets_val = triplets_val[:len_dataset]

    transform = get_normalisation_resize_albu(im_size=32)
    train_dataset = TriDataset(triplets_train, transforms=transform, im_root=MOCK_DATASET_PATH, expand_ratio=0)
    val_dataset = TriDataset(triplets_val, transforms=transform, im_root=MOCK_DATASET_PATH, expand_ratio=0)

    train_dataloader = DataLoader(
        dataset=train_dataset, shuffle=True, drop_last=False, batch_size=batch_size, collate_fn=tri_collate
    )
    val_dataloader = DataLoader(
        dataset=val_dataset, shuffle=False, drop_last=False, batch_size=batch_size, collate_fn=tri_collate
    )

    trainer_engine_params = parse_engine_params_from_config({"accelerator": "cpu", "devices": devices})
    pl_model = DummyModule(exp_num=exp_num, loaders_val=val_dataloader, loaders_train=train_dataloader)

    trainer = Trainer(
        max_epochs=max_epochs,
        enable_checkpointing=False,
        enable_model_summary=False,
        num_sanity_val_steps=0,
        logger=False,
        **trainer_engine_params,
    )
    trainer.fit(model=pl_model)


def get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("--devices", type=int)
    parser.add_argument("--exp_num", type=int)
    parser.add_argument("--len_dataset", type=int)
    parser.add_argument("--batch_size", type=int)
    parser.add_argument("--max_epochs", type=int)
    return parser


if __name__ == "__main__":
    experiment(get_parser().parse_args())



import os
import random
from typing import Any, Dict, Iterable, List, Tuple

import dotenv
import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf

from oml.const import DOTENV_PATH, TCfg


def find_value_ids(it: Iterable[Any], value: Any) -> List[int]:
    """
    Args:
        it: List of any
        value: Query element

    Returns:
        Indices of the all elements equal to x0
    """
    if isinstance(it, np.ndarray):
        inds = list(np.where(it == value)[0])
    else:  # could be very slow
        inds = [i for i, el in enumerate(it) if el == value]
    return inds


def set_global_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

    os.environ["PL_SEED_WORKERS"] = str(1)

    try:
        import torch_xla.core.xla_model as xm
    except ImportError:
        pass
    else:
        xm.set_rng_state(seed)


def one_hot(i: int, dim: int) -> torch.Tensor:
    vector = torch.zeros(dim)
    vector[i] = 1
    return vector


def flatten_dict(
    d: Dict[str, Any], parent_key: str = "", sep: str = "/", ignored_keys: Iterable[str] = ()
) -> Dict[str, Any]:
    items = []  # type: ignore
    for k, v in d.items():
        if k in ignored_keys:
            continue
        new_key = str(parent_key) + sep + str(k) if parent_key else str(k)
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep, ignored_keys=ignored_keys).items())
        else:
            items.append((new_key, v))
    return dict(items)


def load_dotenv() -> None:
    dotenv.load_dotenv(DOTENV_PATH)


def dictconfig_to_dict(cfg: TCfg) -> Dict[str, Any]:
    if isinstance(cfg, DictConfig):
        cfg = OmegaConf.to_container(cfg, resolve=True)

    ret = dict()

    for k in cfg.keys():

        if isinstance(cfg[k], DictConfig) or isinstance(cfg[k], dict):
            ret[k] = dictconfig_to_dict(cfg[k])
        else:
            ret[k] = cfg[k]

    return ret


def smart_sample(array: List[Any], k: int) -> List[Any]:
    """Sample n_samples items from given list. If array contains at least n_samples items, sample without repetition;
    otherwise take all the unique items and sample n_samples - len(array) ones with repetition.

    Args:
        array: list of unique elements to sample from
        k: number of items to sample

    Returns:
        sampled_items: list of sampled items
    """
    array_size = len(array)
    if array_size < k:
        sampled = (
            np.random.choice(array, size=array_size, replace=False).tolist()
            + np.random.choice(array, size=k - array_size, replace=True).tolist()
        )
    else:
        sampled = np.random.choice(array, size=k, replace=False).tolist()
    return sampled


def clip_max(arr: Tuple[int, ...], max_el: int) -> Tuple[int, ...]:
    return tuple(min(x, max_el) for x in arr)


__all__ = [
    "find_value_ids",
    "set_global_seed",
    "one_hot",
    "flatten_dict",
    "load_dotenv",
    "dictconfig_to_dict",
    "smart_sample",
    "clip_max",
]



import math
from collections import defaultdict
from functools import partial
from typing import Any, Tuple

import pytest
import torch

from oml.const import (
    CATEGORIES_KEY,
    EMBEDDINGS_KEY,
    IS_GALLERY_KEY,
    IS_QUERY_KEY,
    LABELS_KEY,
    OVERALL_CATEGORIES_KEY,
    PATHS_KEY,
)
from oml.metrics.embeddings import EmbeddingMetrics
from oml.utils.misc import one_hot

oh = partial(one_hot, dim=8)


def check_dicts_of_dicts_are_equal(dict1: Any, dict2: Any) -> None:
    for key1 in dict1:
        for key2 in dict1[key1]:
            assert dict1[key1][key2] == dict2[key1][key2]

    for key1 in dict2:
        for key2 in dict2[key1]:
            assert dict2[key1][key2] == dict1[key1][key2]


@pytest.fixture()
def perfect_case() -> Any:
    """
    Here we assume that our model provides the best possible embeddings:
    for the item with the label parameter equals to i,
    it provides one hot vector with non zero element in i-th position.

    Thus, we expect all of the metrics equals to 1.
    """

    batch1 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([True, True, True]),
        IS_GALLERY_KEY: torch.tensor([False, False, False]),
        CATEGORIES_KEY: ["cat", "dog", "dog"],
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([False, False, False]),
        IS_GALLERY_KEY: torch.tensor([True, True, True]),
        CATEGORIES_KEY: ["cat", "dog", "dog"],
    }

    k = 1
    metrics = defaultdict(lambda: defaultdict(dict))  # type: ignore
    metrics[OVERALL_CATEGORIES_KEY]["cmc"][k] = 1.0
    metrics["cat"]["cmc"][k] = 1.0
    metrics["dog"]["cmc"][k] = 1.0

    return (batch1, batch2), (metrics, k)


@pytest.fixture()
def imperfect_case() -> Any:
    batch1 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(3)]),  # 3d embedding pretends to be an error
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([True, True, True]),
        IS_GALLERY_KEY: torch.tensor([False, False, False]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([False, False, False]),
        IS_GALLERY_KEY: torch.tensor([True, True, True]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }

    k = 1
    metrics = defaultdict(lambda: defaultdict(dict))  # type: ignore
    metrics[OVERALL_CATEGORIES_KEY]["cmc"][k] = 0.6666666865348816  # it's 2/3 in float precision
    metrics[10]["cmc"][k] = 1.0
    metrics[20]["cmc"][k] = 0.5

    return (batch1, batch2), (metrics, k)


@pytest.fixture()
def worst_case() -> Any:
    batch1 = {
        EMBEDDINGS_KEY: torch.stack([oh(1), oh(0), oh(0)]),  # 3d embedding pretends to be an error
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([True, True, True]),
        IS_GALLERY_KEY: torch.tensor([False, False, False]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([False, False, False]),
        IS_GALLERY_KEY: torch.tensor([True, True, True]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }

    k = 1
    metrics = defaultdict(lambda: defaultdict(dict))  # type: ignore
    metrics[OVERALL_CATEGORIES_KEY]["cmc"][k] = 0
    metrics[10]["cmc"][k] = 0
    metrics[20]["cmc"][k] = 0

    return (batch1, batch2), (metrics, k)


@pytest.fixture()
def case_for_distance_check() -> Any:
    batch1 = {
        EMBEDDINGS_KEY: torch.stack([oh(1) * 2, oh(1) * 3, oh(0)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([True, True, True]),
        IS_GALLERY_KEY: torch.tensor([False, False, False]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([False, False, False]),
        IS_GALLERY_KEY: torch.tensor([True, True, True]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
    }
    ids_ranked_by_distance = [1, 2, 0]
    return (batch1, batch2), ids_ranked_by_distance


def run_retrieval_metrics(case) -> None:  # type: ignore
    (batch1, batch2), (gt_metrics, k) = case

    top_k = (k,)

    num_samples = len(batch1[LABELS_KEY]) + len(batch2[LABELS_KEY])
    calc = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
        categories_key=CATEGORIES_KEY,
        cmc_top_k=top_k,
        precision_top_k=tuple(),
        map_top_k=tuple(),
    )

    calc.setup(num_samples=num_samples)
    calc.update_data(batch1)
    calc.update_data(batch2)

    metrics = calc.compute_metrics()

    check_dicts_of_dicts_are_equal(gt_metrics, metrics)

    # the euclidean distance between any one-hots is always sqrt(2) or 0
    assert torch.isclose(calc.distance_matrix.unique(), torch.tensor([0, math.sqrt(2)])).all()  # type: ignore

    assert (calc.mask_gt.unique() == torch.tensor([0, 1])).all()  # type: ignore
    assert calc.acc.collected_samples == num_samples  # type: ignore


def run_across_epochs(case1, case2) -> None:  # type: ignore
    (batch11, batch12), (gt_metrics1, k1) = case1
    (batch21, batch22), (gt_metrics2, k2) = case2
    assert k1 == k2

    top_k = (k1,)

    calc = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
        categories_key=CATEGORIES_KEY,
        cmc_top_k=top_k,
        precision_top_k=tuple(),
        map_top_k=tuple(),
    )

    def epoch_case(batch_a, batch_b, ground_truth_metrics) -> None:  # type: ignore
        num_samples = len(batch_a[LABELS_KEY]) + len(batch_b[LABELS_KEY])
        calc.setup(num_samples=num_samples)
        calc.update_data(batch_a)
        calc.update_data(batch_b)
        metrics = calc.compute_metrics()

        check_dicts_of_dicts_are_equal(metrics, ground_truth_metrics)

        # the euclidean distance between any one-hots is always sqrt(2) or 0
        assert torch.isclose(calc.distance_matrix.unique(), torch.tensor([0, math.sqrt(2)])).all()  # type: ignore

        assert (calc.mask_gt.unique() == torch.tensor([0, 1])).all()  # type: ignore
        assert calc.acc.collected_samples == num_samples

    # 1st epoch
    epoch_case(batch11, batch12, gt_metrics1)

    # 2nd epoch
    epoch_case(batch21, batch22, gt_metrics2)

    # 3d epoch
    epoch_case(batch11, batch12, gt_metrics1)

    # 4th epoch
    epoch_case(batch21, batch22, gt_metrics2)


def test_perfect_case(perfect_case) -> None:  # type: ignore
    run_retrieval_metrics(perfect_case)


def test_imperfect_case(imperfect_case) -> None:  # type: ignore
    run_retrieval_metrics(imperfect_case)


def test_worst_case(worst_case) -> None:  # type: ignore
    run_retrieval_metrics(worst_case)


def test_mixed_epochs(perfect_case, imperfect_case, worst_case):  # type: ignore
    cases = [perfect_case, imperfect_case, worst_case]
    for case1 in cases:
        for case2 in cases:
            run_across_epochs(case1, case2)


def test_worst_k(case_for_distance_check) -> None:  # type: ignore
    (batch1, batch2), gt_ids = case_for_distance_check

    num_samples = len(batch1[LABELS_KEY]) + len(batch2[LABELS_KEY])
    calc = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
        categories_key=CATEGORIES_KEY,
        cmc_top_k=(),
        precision_top_k=(),
        map_top_k=(2,),
    )

    calc.setup(num_samples=num_samples)
    calc.update_data(batch1)
    calc.update_data(batch2)

    calc.compute_metrics()

    assert calc.get_worst_queries_ids(f"{OVERALL_CATEGORIES_KEY}/map/2", 3) == gt_ids


@pytest.mark.parametrize("extra_keys", [[], [PATHS_KEY], [PATHS_KEY, "a"], ["a"]])
def test_ready_to_vis(extra_keys: Tuple[str, ...]) -> None:  # type: ignore
    calc = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
        categories_key=CATEGORIES_KEY,
        extra_keys=extra_keys,
        cmc_top_k=(1,),
        precision_top_k=(),
        map_top_k=(),
    )

    assert calc.ready_to_visualize() or PATHS_KEY not in extra_keys



import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Dict, List

import pytest
import torch
from torch.utils.data import SequentialSampler

from oml.const import OVERALL_CATEGORIES_KEY, PROJECT_ROOT
from oml.lightning.modules.module_ddp import ModuleDDP
from oml.samplers.balance import BalanceSampler

from .run_retrieval_experiment_ddp import MetricValCallbackWithSaving

rf"""
MOTIVATION

With this experiment, we want to test the patching of loaders with {ModuleDDP} and the similarity of metrics
in DDP mode.

We check the following:
1) Train and Val loaders are split into several parts. These parts have no overlapping except for several samples (for
validation with default {SequentialSampler}) or several batches (for training with {BalanceSampler}) which is necessary
for padding according to the number of devices.
2) Metrics obtained with a different number of devices are very similar. For this purpose, we save metrics and
compare them later. Note that only the final metric should be similar.

Our dummy data is presented by GT labels and PRED labels with some errors. Amount of errors is the same for each
runnings.
"""


exp_file = PROJECT_ROOT / "tests/test_examples/test_ddp_cases/run_retrieval_experiment_ddp.py"


@pytest.mark.parametrize("batch_size", [10, 19])
@pytest.mark.parametrize("max_epochs", [2])
@pytest.mark.parametrize("num_labels,atol", [(200, 5e-3), (1000, 2e-2)])
def test_metrics_is_similar_in_ddp(num_labels: int, atol: float, batch_size: int, max_epochs: int) -> None:
    devices = (1, 2, 3)
    # We will compare metrics from same experiment but with different amount of devices. For this we aggregate
    # metrics in variable with following structure:
    # {"cmc_<K>": [cmc_<K>_<exp_1>, cmc_<K>_<exp_2>, cmc_<K>_<exp_3>, ...],
    #  "map_<K>": [map_<K>_<exp_1>, map_<K>_<exp_2>, map_<K>_<exp_3>, ...]}
    # After that we will compare metrics. Metrics shouldn't be equal, but very similar
    metric_topk2values = defaultdict(list)

    for num_devices in devices:
        params = (
            f"--devices {num_devices} "
            f"--max_epochs {max_epochs} "
            f"--num_labels {num_labels} "
            f"--batch_size {batch_size}"
        )
        cmd = f"python {exp_file} " + params
        subprocess.run(cmd, check=True, shell=True)

        metrics_path = MetricValCallbackWithSaving.save_path_pattern.format(
            devices=num_devices, batch_size=batch_size, num_labels=num_labels
        )
        metrics = torch.load(metrics_path)[OVERALL_CATEGORIES_KEY]
        Path(metrics_path).unlink(missing_ok=True)
        for metric_name, topk2value in metrics.items():
            for top_k, value in topk2value.items():
                metric_topk2values[f"{metric_name}_{top_k}"].append(value)

    compare_metrics(metric_topk2values, atol)


def compare_metrics(metric_topk2values: Dict[str, List[torch.Tensor]], atol: float) -> None:
    # Check point 2 of motivation
    for metric_topk, values in metric_topk2values.items():
        mean_value = torch.tensor(values).float().mean()
        is_close = tuple(torch.isclose(val, mean_value, atol=atol) for val in values)
        assert all(is_close), f"Metrics are not similar: {[metric_topk, values, is_close, atol]}"



import math
from random import randint, shuffle
from typing import Any, Dict, Optional

import numpy as np
import pytest
import torch
from torch.utils.data import DataLoader

from oml.const import INPUT_TENSORS_KEY, LABELS_KEY
from oml.interfaces.datasets import IDatasetWithLabels
from oml.losses.triplet import TripletLossWithMiner
from oml.miners.cross_batch import TripletMinerWithMemory
from oml.registry.miners import get_miner
from oml.samplers.balance import BalanceSampler
from tests.test_integrations.utils import IdealOneHotModel


class DummyDataset(IDatasetWithLabels):
    def __init__(self, n_labels: int, n_samples_min: int):
        self.labels = []
        for i in range(n_labels):
            self.labels.extend([i] * randint(n_samples_min, 2 * n_samples_min))
        shuffle(self.labels)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        return {INPUT_TENSORS_KEY: torch.tensor(self.labels[idx]), LABELS_KEY: self.labels[idx]}

    def __len__(self) -> int:
        return len(self.labels)

    def get_labels(self) -> np.ndarray:
        return np.array(self.labels)


@pytest.mark.parametrize(
    "miner_name,miner_params",
    [
        ("hard_triplets", {"norm_required": True}),
        ("all_triplets", dict()),
        ("hard_cluster", dict()),
        ("triplets_with_memory", {"bank_size_in_batches": 5, "tri_expand_k": 3}),
    ],
)
@pytest.mark.parametrize("margin", [None, 0.5])
@pytest.mark.parametrize("n_labels,n_instances", [(2, 2), (5, 6)])
def test_train_with_mining(
    miner_name: str,
    miner_params: Dict[str, Any],
    margin: Optional[float],
    n_labels: int,
    n_instances: int,
) -> None:
    n_labels_total = n_labels * 6  # just some random figures

    dataset = DummyDataset(n_labels=n_labels_total, n_samples_min=n_instances)

    model = IdealOneHotModel(emb_dim=n_labels_total + randint(1, 5))

    loader = DataLoader(
        dataset=dataset,
        batch_sampler=BalanceSampler(labels=dataset.get_labels(), n_labels=n_labels, n_instances=n_instances),
    )

    miner = get_miner(miner_name, **miner_params)
    criterion = TripletLossWithMiner(margin=margin, miner=miner, need_logs=False)

    for i, batch in enumerate(loader):
        assert len(batch[LABELS_KEY]) == n_labels * n_instances

        embeddings = model(batch[INPUT_TENSORS_KEY])
        loss = criterion(embeddings, batch[LABELS_KEY])

        if isinstance(miner, TripletMinerWithMemory) and (i < miner.bank_size_in_batches):
            # we cannot guarantee any values of loss due to impact of memory bank initialisation
            continue
        else:
            if margin is None:
                # soft triplet loss
                assert loss.isclose(torch.log1p(torch.exp(torch.tensor(-math.sqrt(2)))), atol=1e-6)
            else:
                assert loss.isclose(torch.tensor(0.0))



import torch

from oml.const import INPUT_TENSORS_KEY
from oml.datasets.triplet import tri_collate
from oml.losses.triplet import get_tri_ids_in_plain


def test_collate() -> None:
    # test absence of shuffle after collate function
    def fdata(x: int) -> torch.tensor:
        return x * torch.ones(3, 4, 4)

    # order is: anchor, positive, negative
    items = [
        {INPUT_TENSORS_KEY: (fdata(1), fdata(2), fdata(3))},
        {INPUT_TENSORS_KEY: (fdata(11), fdata(22), fdata(33))},
        {INPUT_TENSORS_KEY: (fdata(111), fdata(222), fdata(333))},
    ]

    batch = tri_collate(items)[INPUT_TENSORS_KEY]

    ii_a, ii_p, ii_n = get_tri_ids_in_plain(n=len(batch))

    assert (batch[ii_a] % 10 == 1).all()
    assert (batch[ii_p] % 10 == 2).all()
    assert (batch[ii_n] % 10 == 3).all()



from typing import Dict, List

import pytest
import torch
from pytorch_lightning.plugins import DDPPlugin

from oml.const import TCfg
from oml.lightning.entrypoints.parser import parse_engine_params_from_config


def compare_params(cfg: TCfg, expected_params: TCfg) -> None:
    parsed_params = parse_engine_params_from_config(cfg)

    assert set(parsed_params.keys()) == set(expected_params.keys())

    for key in expected_params.keys():
        if key == "strategy":
            assert type(expected_params[key]) == type(parsed_params[key])
        else:

            assert expected_params[key] == parsed_params[key]


@pytest.mark.parametrize("list_devices", [[1], [1, 2], [0, 3], [0, 1, 2, 5]])
def test_list_devices_cpu_accelerator(list_devices: List[int]) -> None:
    cfg = {"accelerator": "cpu", "devices": list_devices}
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "cpu",
        "devices": len(list_devices),
        "strategy": DDPPlugin() if len(list_devices) > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("int_devices", [1, 2, 4])
def test_int_devices_cpu_accelerator(int_devices: int) -> None:
    cfg = {"accelerator": "cpu", "devices": int_devices}
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "cpu",
        "devices": int_devices,
        "strategy": DDPPlugin() if int_devices > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("list_devices", [[1], [1, 2], [0, 3], [0, 1, 2, 5]])
def test_list_devices_gpu_accelerator(list_devices: List[int]) -> None:
    cfg = {"accelerator": "gpu", "devices": list_devices}
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "gpu",
        "devices": list_devices,
        "strategy": DDPPlugin() if len(list_devices) > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("int_devices", [1, 2, 4])
def test_int_devices_gpu_accelerator(int_devices: int) -> None:
    cfg = {"accelerator": "gpu", "devices": int_devices}
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "gpu",
        "devices": int_devices,
        "strategy": DDPPlugin() if int_devices > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("no_devices", [{}, {"devices": None}])
@pytest.mark.parametrize("no_accelerator", [{}, {"accelerator": None}])
def test_no_devices_no_accelerator(no_devices: TCfg, no_accelerator: TCfg) -> None:
    cfg = {}  # type: ignore
    cfg.update(no_devices)
    cfg.update(no_accelerator)

    if torch.cuda.is_available():
        accelerator = "gpu"
        devices = torch.cuda.device_count()
        if devices > 1:
            strategy = DDPPlugin()
        else:
            strategy = None
    else:
        accelerator = "cpu"
        devices = 1
        strategy = None

    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": accelerator,
        "devices": devices,
        "strategy": strategy,
    }

    compare_params(cfg, expected_params)


@pytest.mark.parametrize("no_devices", [{}, {"devices": None}])
def test_no_devices_cpu_accelerator(no_devices: TCfg) -> None:
    cfg = {"accelerator": "cpu"}
    cfg.update(no_devices)

    expected_params = {"gpus": None, "replace_sampler_ddp": False, "accelerator": "cpu", "devices": 1, "strategy": None}
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("no_devices", [{}, {"devices": None}])
def test_no_devices_gpu_accelerator(no_devices: TCfg) -> None:
    cfg = {"accelerator": "gpu"}
    cfg.update(no_devices)

    devices = torch.cuda.device_count() if torch.cuda.is_available() else 1
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "gpu",
        "devices": devices,
        "strategy": DDPPlugin() if devices > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("no_accelerator", [{}, {"accelerator": None}])
@pytest.mark.parametrize("int_devices", [1, 2, 4])
def case_int_devices_no_accelerator(no_accelerator: TCfg, int_devices: int) -> None:
    cfg = {"devices": int_devices}
    cfg.update(no_accelerator)

    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": "gpu" if torch.cuda.is_available() else "cpu",
        "devices": int_devices,
        "strategy": DDPPlugin() if int_devices > 1 else None,
    }
    compare_params(cfg, expected_params)


@pytest.mark.parametrize("no_accelerator", [{}, {"accelerator": None}])
@pytest.mark.parametrize("list_devices", [[1], [1, 2], [0, 3], [0, 1, 2, 5]])
def case_list_devices_no_accelerator(no_accelerator: TCfg, list_devices: List[int]) -> None:
    cfg = {"devices": list_devices}
    cfg.update(no_accelerator)

    accelerator = "gpu" if torch.cuda.is_available() else "cpu"
    devices = len(list_devices) if accelerator == "cpu" else list_devices
    expected_params = {
        "gpus": None,
        "replace_sampler_ddp": False,
        "accelerator": accelerator,
        "devices": devices,
        "strategy": DDPPlugin() if len(list_devices) > 1 else None,
    }
    compare_params(cfg, expected_params)



from collections import defaultdict
from functools import partial
from typing import Any

import pytest
import torch

from oml.const import CATEGORIES_KEY, EMBEDDINGS_KEY, OVERALL_CATEGORIES_KEY
from oml.metrics.triplets import AccuracyOnTriplets
from oml.utils.misc import one_hot

oh = partial(one_hot, dim=8)


@pytest.fixture()
def perfect_case() -> Any:
    batch = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(0), oh(1), oh(1), oh(1), oh(0)]),  # 1st and 2nd triplet
        CATEGORIES_KEY: ["cat", "dog"],  # 1st triplet  # 2nd triplet
    }

    gt_metrics = defaultdict(dict)  # type: ignore
    gt_metrics[OVERALL_CATEGORIES_KEY][AccuracyOnTriplets.metric_name] = 1
    gt_metrics["cat"][AccuracyOnTriplets.metric_name] = 1
    gt_metrics["dog"][AccuracyOnTriplets.metric_name] = 1

    return [batch], gt_metrics


@pytest.fixture()
def some_case() -> Any:
    batch1 = {
        EMBEDDINGS_KEY: torch.stack(
            [
                # triplet #1 - error
                oh(0),
                oh(5),
                oh(0),
                # triplet #2 - error
                oh(1),
                oh(0),
                oh(1),
                # triplet #3 - correct
                oh(5),
                oh(5),
                oh(1),
                # triplet #4 - correct
                oh(3),
                oh(3),
                oh(2),
            ]
        ),
        CATEGORIES_KEY: [0, 1, 1, 1],  # triplets #1 #2 #3 #4
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack(
            [
                # triplet #5 - correct
                oh(0),
                oh(0),
                oh(1),
                # triplet #6 - error
                oh(1),
                oh(0),
                oh(1),
            ]
        ),
        CATEGORIES_KEY: [0, 1],  # triplets #5 #6
    }

    gt_metrics = defaultdict(dict)  # type: ignore
    gt_metrics[OVERALL_CATEGORIES_KEY][AccuracyOnTriplets.metric_name] = 1 / 2
    gt_metrics[0][AccuracyOnTriplets.metric_name] = 1 / 2
    gt_metrics[1][AccuracyOnTriplets.metric_name] = 1 / 2

    return [batch1, batch2], gt_metrics


def run_accuracy_on_triplets(case) -> None:  # type: ignore
    batches, gt_metrics = case

    acc = AccuracyOnTriplets(embeddings_key=EMBEDDINGS_KEY, categories_key=CATEGORIES_KEY)
    acc.setup()
    for batch in batches:
        acc.update_data(batch)

    metrics = acc.compute_metrics()

    assert gt_metrics == metrics, (gt_metrics, metrics)


def test_perfect_case(perfect_case) -> None:  # type: ignore
    run_accuracy_on_triplets(perfect_case)


def test_some_case(some_case) -> None:  # type: ignore
    run_accuracy_on_triplets(some_case)



import os
from pathlib import Path
from pprint import pprint

import albumentations as albu
import pytorch_lightning as pl
from pytorch_lightning.loggers import NeptuneLogger
from torch.utils.data import DataLoader

from oml.const import (
    CATEGORIES_COLUMN,
    LABELS_COLUMN,
    OVERALL_CATEGORIES_KEY,
    PROJECT_ROOT,
    TCfg,
)
from oml.datasets.base import get_retrieval_datasets
from oml.interfaces.criterions import ITripletLossWithMiner
from oml.interfaces.models import IExtractor
from oml.lightning.callbacks.metric import MetricValCallback, MetricValCallbackDDP
from oml.lightning.entrypoints.parser import (
    check_is_config_for_ddp,
    parse_engine_params_from_config,
)
from oml.lightning.modules.retrieval import RetrievalModule, RetrievalModuleDDP
from oml.metrics.embeddings import EmbeddingMetrics, EmbeddingMetricsDDP
from oml.registry.losses import get_criterion_by_cfg
from oml.registry.models import get_extractor_by_cfg
from oml.registry.optimizers import get_optimizer_by_cfg
from oml.registry.samplers import SAMPLERS_CATEGORIES_BASED, get_sampler_by_cfg
from oml.registry.schedulers import get_scheduler_by_cfg
from oml.registry.transforms import get_transforms_by_cfg
from oml.utils.misc import (
    dictconfig_to_dict,
    flatten_dict,
    load_dotenv,
    set_global_seed,
)


def pl_train(cfg: TCfg) -> None:
    """
    This is an entrypoint for the model training in metric learning setup.

    The config can be specified as a dictionary or with hydra: https://hydra.cc/.
    For more details look at ``examples/README.md``

    """
    cfg = dictconfig_to_dict(cfg)
    trainer_engine_params = parse_engine_params_from_config(cfg)
    is_ddp = check_is_config_for_ddp(trainer_engine_params)

    pprint(cfg)

    set_global_seed(cfg["seed"])

    cwd = Path.cwd()

    transforms_train = get_transforms_by_cfg(cfg["transforms_train"])
    transforms_val = get_transforms_by_cfg(cfg["transforms_val"])

    train_dataset, valid_dataset = get_retrieval_datasets(
        dataset_root=Path(cfg["dataset_root"]),
        transforms_train=transforms_train,
        transforms_val=transforms_val,
        dataframe_name=cfg["dataframe_name"],
        cache_size=cfg["cache_size"],
    )

    if isinstance(transforms_train, albu.Compose):
        augs_file = ".hydra/augs_cfg.yaml" if Path(".hydra").exists() else "augs_cfg.yaml"
        albu.save(filepath=augs_file, transform=transforms_train, data_format="yaml")
    else:
        augs_file = None

    if (not train_dataset.categories_key) and cfg["sampler"]["name"] in SAMPLERS_CATEGORIES_BASED.keys():
        raise ValueError(
            "NOTE! You are trying to use Sampler which works with the information related"
            "to categories, but there is no <categories_key> in your Dataset."
        )

    sampler_runtime_args = {"labels": train_dataset.get_labels()}
    if train_dataset.categories_key:
        df = train_dataset.df
        sampler_runtime_args["label2category"] = dict(zip(df[LABELS_COLUMN], df[CATEGORIES_COLUMN]))
    # note, we pass some runtime arguments to sampler here, but not all of the samplers use all of these arguments
    sampler = get_sampler_by_cfg(cfg["sampler"], **sampler_runtime_args) if cfg["sampler"] is not None else None

    extractor = get_extractor_by_cfg(cfg["model"])
    criterion = get_criterion_by_cfg(cfg["criterion"])
    optimizer = get_optimizer_by_cfg(cfg["optimizer"], params=extractor.parameters())

    # unpack scheduler to the Lightning format
    if cfg.get("scheduling"):
        scheduler_kwargs = {
            "scheduler": get_scheduler_by_cfg(cfg["scheduling"]["scheduler"], optimizer=optimizer),
            "scheduler_interval": cfg["scheduling"]["scheduler_interval"],
            "scheduler_frequency": cfg["scheduling"]["scheduler_frequency"],
            "scheduler_monitor_metric": cfg["scheduling"].get("monitor_metric", None),
        }
    else:
        scheduler_kwargs = {"scheduler": None}

    assert isinstance(extractor, IExtractor), "You model must to be child of IExtractor"

    if sampler is None:
        loader_train = DataLoader(
            dataset=train_dataset,
            num_workers=cfg["num_workers"],
            batch_size=cfg["bs_train"],
            drop_last=True,
            shuffle=True,
        )
    else:
        loader_train = DataLoader(
            dataset=train_dataset,
            batch_sampler=sampler,
            num_workers=cfg["num_workers"],
        )

    loaders_val = DataLoader(dataset=valid_dataset, batch_size=cfg["bs_val"], num_workers=cfg["num_workers"])

    module_kwargs = scheduler_kwargs
    if is_ddp:
        module_kwargs.update({"loaders_train": loader_train, "loaders_val": loaders_val})
        module_constructor = RetrievalModuleDDP
    else:
        module_constructor = RetrievalModule  # type: ignore

    pl_model = module_constructor(
        model=extractor,
        criterion=criterion,
        optimizer=optimizer,
        input_tensors_key=train_dataset.input_tensors_key,
        labels_key=train_dataset.labels_key,
        **module_kwargs,
    )

    metrics_constructor = EmbeddingMetricsDDP if is_ddp else EmbeddingMetrics
    metrics_calc = metrics_constructor(
        embeddings_key=pl_model.embeddings_key,
        categories_key=valid_dataset.categories_key,
        labels_key=valid_dataset.labels_key,
        is_query_key=valid_dataset.is_query_key,
        is_gallery_key=valid_dataset.is_gallery_key,
        extra_keys=(valid_dataset.paths_key, *valid_dataset.bboxes_keys),
        **cfg.get("metric_args", {}),
    )

    metrics_clb_constructor = MetricValCallbackDDP if is_ddp else MetricValCallback
    metrics_clb = metrics_clb_constructor(
        metric=metrics_calc,
        log_images=cfg.get("log_images", False),
    )
    ckpt_clb = pl.callbacks.ModelCheckpoint(
        dirpath=Path.cwd() / "checkpoints",
        monitor=cfg.get("metric_for_checkpointing", f"{OVERALL_CATEGORIES_KEY}/cmc/1"),
        mode="max",
        save_top_k=1,
        verbose=True,
        filename="best",
    )

    # Here we try to load NEPTUNE_API_TOKEN from .env file
    # You can also set it up via `export NEPTUNE_API_TOKEN=...`
    load_dotenv()
    if ("NEPTUNE_API_TOKEN" in os.environ.keys()) and (cfg["neptune_project"] is not None):
        logger = NeptuneLogger(
            api_key=os.environ["NEPTUNE_API_TOKEN"],
            project=cfg["neptune_project"],
            tags=list(cfg["tags"]) + [cfg["postfix"]] + [cwd.name],
            log_model_checkpoints=False,
        )
        # log hyper params and files
        dict_to_log = {**dictconfig_to_dict(cfg), **{"dir": cwd}}
        logger.log_hyperparams(flatten_dict(dict_to_log, sep="|"))
        logger.run["dataset"].upload(str(Path(cfg["dataset_root"]) / cfg["dataframe_name"]))
        if augs_file is not None:
            logger.run["augs_cfg"].upload(augs_file)
        # log source code
        source_files = list(map(lambda x: str(x), PROJECT_ROOT.glob("**/*.py"))) + list(
            map(lambda x: str(x), PROJECT_ROOT.glob("**/*.yaml"))
        )
        logger.run["code"].upload_files(source_files)

    else:
        logger = True

    trainer = pl.Trainer(
        max_epochs=cfg["max_epochs"],
        num_sanity_val_steps=0,
        check_val_every_n_epoch=cfg["valid_period"],
        default_root_dir=cwd,
        enable_checkpointing=True,
        enable_progress_bar=True,
        enable_model_summary=True,
        callbacks=[metrics_clb, ckpt_clb],
        logger=logger,
        precision=cfg.get("precision", 32),
        **trainer_engine_params,
    )

    if is_ddp:
        trainer.fit(model=pl_model)
    else:
        trainer.fit(model=pl_model, train_dataloaders=loader_train, val_dataloaders=loaders_val)


__all__ = ["pl_train"]



import hydra
import torchvision.transforms as t
from omegaconf import DictConfig

from oml.const import MOCK_DATASET_PATH
from oml.lightning.entrypoints.train import pl_train
from oml.registry.transforms import TRANSFORMS_REGISTRY
from oml.utils.download_mock_dataset import download_mock_dataset
from oml.utils.misc import dictconfig_to_dict


def get_custom_augs(im_size: int) -> t.Compose:
    return t.Compose(
        [
            t.Resize(size=(im_size, im_size)),
            t.RandomHorizontalFlip(),
            t.RandomGrayscale(),
            t.ToTensor(),
            t.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ]
    )


TRANSFORMS_REGISTRY["custom_augmentations"] = get_custom_augs  # type: ignore


@hydra.main(config_path="configs", config_name="train_with_bboxes.yaml")
def main_hydra(cfg: DictConfig) -> None:
    cfg = dictconfig_to_dict(cfg)
    download_mock_dataset(MOCK_DATASET_PATH)
    cfg["dataset_root"] = MOCK_DATASET_PATH
    pl_train(cfg)


if __name__ == "__main__":
    main_hydra()



import shutil
import subprocess
import warnings
from pathlib import Path
from typing import List, Tuple

import pytest
import torch
import yaml  # type: ignore

from oml.const import PROJECT_ROOT

warnings.filterwarnings("ignore")

SCRIPTS_PATH = PROJECT_ROOT / "tests/test_examples/test_runs_via_configs/"


def accelerator_devices_pairs() -> List[Tuple[str, int]]:
    pairs = [("cpu", 1), ("cpu", 2)]

    if torch.cuda.is_available():
        pairs += [("gpu", 1)]
        if torch.cuda.device_count() > 1:
            pairs += [("gpu", 2)]

    return pairs


def rm_logs(cfg_name: Path) -> None:
    with open(SCRIPTS_PATH / "configs" / cfg_name, "r") as f:
        cfg = yaml.safe_load(f)

    if ("logs_root" in cfg) and Path(cfg["logs_root"]).exists():
        shutil.rmtree(cfg["logs_root"])


def run(file: str, accelerator: str, devices: int) -> None:
    cmd = f"python {str(SCRIPTS_PATH / file)} ++accelerator='{accelerator}' ++devices='{devices}'"
    subprocess.run(cmd, check=True, shell=True)

    rm_logs(cfg_name=SCRIPTS_PATH / "configs" / file.replace(".py", ".yaml"))


@pytest.mark.parametrize("accelerator, devices", accelerator_devices_pairs())
def test_train(accelerator: str, devices: int) -> None:
    run("train.py", accelerator, devices)


@pytest.mark.parametrize("accelerator, devices", accelerator_devices_pairs())
def test_train_with_bboxes(accelerator: str, devices: int) -> None:
    run("train_with_bboxes.py", accelerator, devices)


@pytest.mark.parametrize("accelerator, devices", accelerator_devices_pairs())
def test_train_with_categories(accelerator: str, devices: int) -> None:
    run("train_with_categories.py", accelerator, devices)


@pytest.mark.parametrize("accelerator, devices", accelerator_devices_pairs())
def test_val(accelerator: str, devices: int) -> None:
    run("val_mock.py", accelerator, devices)



from collections import Counter
from operator import itemgetter
from random import randint, shuffle
from typing import List, Set

import pytest

from oml.samplers.balance import BalanceSampler

TLabels = List[List[int]]


def generate_valid_labels(num: int) -> TLabels:
    """
    This function generates some valid inputs for samplers.
    It generates n_instances for n_labels.

    Args:
        num: Number of generated samples

    Returns:
        Labels

    """
    labels_generated = []

    for _ in range(num):
        n_labels = randint(2, 30)
        labels_list = [[label] * randint(2, 10) for label in range(n_labels)]
        labels = [el for sublist in labels_list for el in sublist]

        shuffle(labels)
        labels_generated.append(labels)

    return labels_generated


@pytest.fixture()
def input_for_balance_batch_sampler() -> TLabels:
    """
    Returns:
        Test data for sampler in the following order: (labels, n_labels, n_instances)

    """
    input_cases = [
        # ideal case
        [0, 1, 2, 3, 0, 1, 2, 3],
        # repetition sampling is needed for label #3
        [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2],
        # check last batch behaviour:
        # last batch includes less than n_labels (2 < 3)
        [0, 1, 2, 3, 4, 0, 1, 2, 3, 4],
        # we need to drop 1 label during the epoch because
        # number of labels in data % n_labels = 1
        [0, 1, 2, 3, 0, 1, 2, 3],
        # several random cases
        [0, 1, 2, 2, 1, 0, 1, 0, 2, 0, 1, 2],
        [0, 1, 2, 2, 1, 0, 1, 0, 2, 0, 1, 2],
        [0, 1, 2, 2, 1, 0, 1, 0, 2, 0, 1, 2],
    ]

    # (alekseysh) It was checked once with N = 100_000 before doing the PR
    num_random_cases = 100
    input_cases.extend((generate_valid_labels(num_random_cases)))

    return input_cases


def check_balance_batch_sampler_epoch(sampler: BalanceSampler, labels: List[int]) -> None:
    sampled_ids = list(sampler)
    collected_labels: Set[int] = set()

    # emulating of 1 epoch
    for i, batch_ids in enumerate(sampled_ids):
        batch_labels = itemgetter(*batch_ids)(labels)  # type: ignore

        labels_counter = Counter(batch_labels)
        num_batch_labels = len(labels_counter)
        num_batch_samples_counts = list(labels_counter.values())

        # batch-level invariants
        assert all(label not in collected_labels for label in batch_labels)
        assert len(set(batch_ids)) >= 4, set(batch_ids)  # type: ignore
        assert len(batch_ids) == sampler.n_labels * sampler.n_instances
        assert num_batch_labels == sampler.n_labels, (num_batch_labels, sampler.n_labels)
        assert all(el == sampler.n_instances for el in num_batch_samples_counts)

        collected_labels.update(batch_labels)

    # epoch-level invariants
    n_expected_batches = len(set(labels)) // sampler.n_labels
    num_labels_in_data = len(set(labels))
    num_labels_sampled = len(collected_labels)
    n_instances_sampled = sum(map(len, sampled_ids))  # type: ignore

    assert len(sampler) == n_expected_batches, (len(sampler), n_expected_batches)
    assert (num_labels_in_data - sampler.n_labels + 1) <= num_labels_sampled <= num_labels_in_data
    assert len(sampler) * sampler.n_labels * sampler.n_instances == n_instances_sampled


def test_balance_batch_sampler(input_for_balance_batch_sampler: TLabels) -> None:
    for labels in input_for_balance_batch_sampler:
        n_labels_batch = randint(2, max(2, len(set(labels)) // 5))
        n_instances_batch = randint(2, max(Counter(labels).values()))
        sampler = BalanceSampler(labels=labels, n_labels=n_labels_batch, n_instances=n_instances_batch)
        check_balance_batch_sampler_epoch(sampler=sampler, labels=labels)



from math import isclose
from typing import Any, Dict, Tuple

import pytest
import torch
from torch import Tensor
from torch.utils.data import DataLoader

from oml.const import (
    EMBEDDINGS_KEY,
    INPUT_TENSORS_KEY,
    IS_GALLERY_KEY,
    IS_QUERY_KEY,
    LABELS_KEY,
    OVERALL_CATEGORIES_KEY,
)
from oml.interfaces.datasets import IDatasetQueryGallery
from oml.metrics.embeddings import EmbeddingMetrics
from tests.test_integrations.utils import IdealClusterEncoder

TData = Tuple[Tensor, Tensor, Tensor, Tensor, float]


def get_separate_query_gallery() -> TData:
    labels = torch.tensor([0, 0, 1, 1, 2, 2, 2, 2, 1])
    gallery_mask = torch.tensor([0, 1, 0, 1, 0, 0, 1, 1, 0])
    query_mask = torch.logical_not(gallery_mask)

    # let's add some errors (swap for some labels) in queries
    input_tensors = labels.clone()
    input_tensors[4] = 0
    input_tensors[5] = 0

    cmc_gt = 3 / 5

    return labels, query_mask, gallery_mask, input_tensors, cmc_gt


def get_shared_query_gallery() -> TData:
    labels = torch.tensor([0, 0, 0, 1, 1, 1, 2, 2]).long()
    query_mask = torch.ones_like(labels)
    gallery_mask = torch.ones_like(labels)

    input_tensors = labels.clone()
    input_tensors[0] = 3

    cmc_gt = 7 / 8

    return labels, query_mask, gallery_mask, input_tensors, cmc_gt


class DummyQGDataset(IDatasetQueryGallery):
    def __init__(self, labels: Tensor, gallery_mask: Tensor, query_mask: Tensor, input_tensors: Tensor):
        assert len(labels) == len(gallery_mask) == len(query_mask)

        self.labels = labels
        self.gallery_mask = gallery_mask
        self.query_mask = query_mask
        self.input_tensors = input_tensors

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        return {
            LABELS_KEY: self.labels[idx],
            INPUT_TENSORS_KEY: self.input_tensors[idx],
            IS_QUERY_KEY: self.query_mask[idx],
            IS_GALLERY_KEY: self.gallery_mask[idx],
        }

    def __len__(self) -> int:
        return len(self.labels)


@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("shuffle", [True, False])
@pytest.mark.parametrize("num_workers", [0, 2])
@pytest.mark.parametrize("data", [get_separate_query_gallery(), get_shared_query_gallery()])
def test_retrieval_validation(batch_size: int, shuffle: bool, num_workers: int, data: TData) -> None:
    labels, query_mask, gallery_mask, input_tensors, cmc_gt = data

    dataset = DummyQGDataset(
        labels=labels,
        input_tensors=input_tensors,
        query_mask=query_mask,
        gallery_mask=gallery_mask,
    )

    loader = DataLoader(
        dataset=dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, num_workers=num_workers
    )

    calc = EmbeddingMetrics(cmc_top_k=(1,))
    calc.setup(num_samples=len(dataset))

    model = IdealClusterEncoder()

    for batch in loader:
        output = model(batch[INPUT_TENSORS_KEY])
        batch[EMBEDDINGS_KEY] = output
        calc.update_data(data_dict=batch)

    metrics = calc.compute_metrics()

    assert isclose(metrics[OVERALL_CATEGORIES_KEY]["cmc"][1], cmc_gt, abs_tol=1e-5)



from typing import Any, Dict

import torch
from pytorch_lightning.plugins import DDPPlugin

from oml.const import TCfg
from oml.utils.misc import dictconfig_to_dict


def parse_engine_params_from_config(cfg: TCfg) -> Dict[str, Any]:
    """
    The function parses config and based on it prepares DDP parameters for PytorchLightning Trainer module.
    There are two parameters 'accelerator' and 'devices' you can configure. If one of the parameters or both of them
    are not specified, the best option based on your hardware will be automatically prepared for you.
    Possible options for 'accelerator' are 'cpu' and 'gpu'.
    You can select specific GPUs using the list 'devices=[0, 2]' or specify the number of GPUs by `devices=3`.
    An experiment might be launched in DDP with the 'cpu' accelerator. In this case, 'devices' (integer value or
    length of list) interpreted as a number of processes.
    """
    cfg = dictconfig_to_dict(cfg)

    # we want to replace possible null or no values in config for "accelerator" and "devices"
    accelerator = cfg.get("accelerator")
    if accelerator is None:
        accelerator = "gpu" if torch.cuda.is_available() else "cpu"

    devices = cfg.get("devices")
    if devices is None:
        devices = torch.cuda.device_count() if (torch.cuda.is_available() and accelerator == "gpu") else 1

    if isinstance(devices, (list, tuple)) and accelerator == "cpu":
        devices = len(devices)

    if (isinstance(devices, int) and devices > 1) or (isinstance(devices, (list, tuple)) and len(devices) > 1):
        strategy = DDPPlugin()
    else:
        strategy = None

    return {
        "devices": devices,
        "strategy": strategy,
        "accelerator": accelerator,
        "gpus": None,
        "replace_sampler_ddp": False,
    }


def check_is_config_for_ddp(cfg: TCfg) -> bool:
    return bool(cfg["strategy"])


__all__ = ["parse_engine_params_from_config", "check_is_config_for_ddp"]



from random import shuffle

import pytest
import torch
from torch import Tensor, tensor

from oml.miners.cross_batch import TripletMinerWithMemory
from oml.samplers.balance import BalanceSampler
from tests.test_integrations.utils import IdealOneHotModel


def n_possible_triplets(n_labels: int, n_instances: int) -> int:
    return n_labels * (n_labels - 1) * (n_instances - 1) * n_instances**2


def get_labels(n_cls: int, sz: int) -> Tensor:
    labels = []
    for i in range(n_cls):
        labels.extend([i] * sz)
    shuffle(labels)
    return tensor(labels)


@pytest.mark.parametrize("n_cls,cls_sz,n_labels,n_instances", [[100, 8, 2, 8], [50, 10, 10, 4]])
@pytest.mark.parametrize("bank_sz,bank_k", [[5, 3], [20, 5], [5, 1]])
def test_mining_with_memory(
    n_cls: int, cls_sz: int, n_labels: int, n_instances: int, bank_sz: int, bank_k: int
) -> None:
    feat_dim = 2 * n_cls

    labels = get_labels(n_cls=n_cls, sz=cls_sz)
    sampler = BalanceSampler(labels=labels.tolist(), n_labels=n_labels, n_instances=n_instances)

    miner = TripletMinerWithMemory(bank_size_in_batches=bank_sz, tri_expand_k=bank_k)

    # we use shifted one-hot encoding here as a hack to avoid collapses with the features
    # from one-hot initialised memory bank in miner
    model = IdealOneHotModel(emb_dim=feat_dim, shift=bank_k + (n_labels * n_instances + 5))

    n_epoch = 2
    for i_epoch in range(n_epoch):
        for i, ii_batch in enumerate(sampler):
            if len(ii_batch) != n_labels * n_instances:  # to drop_last
                continue

            labels_batch = labels[ii_batch]
            features_batch = model(labels_batch)

            anch, pos, neg, is_original_tri = miner.sample(labels=labels_batch, features=features_batch)
            assert len(anch) == len(pos) == len(neg)
            assert n_possible_triplets(n_labels, n_instances) == int(is_original_tri.sum())

            n_out_tri = len(anch)
            n_desired_tri = bank_k * n_possible_triplets(n_labels, n_instances)

            # during the 1st epoch we have no chance to filter out some of the irrelevant triplets
            # because memory bank is empty
            if (i_epoch == 0) or (bank_k == 1):
                assert n_out_tri == n_desired_tri, (n_out_tri, n_desired_tri)
            # but later some of the collisions are possible, but their amount should not be high
            else:
                assert n_out_tri > 0.9 * n_desired_tri, (n_out_tri, n_desired_tri)

            # we use the fact that our features are one-hot encoded labels
            labels_sampled_anchor = torch.argmax(anch, dim=1)
            labels_sampled_pos = torch.argmax(pos, dim=1)
            labels_sampled_neg = torch.argmax(neg, dim=1)

            assert (labels_sampled_anchor == labels_sampled_pos).all()
            assert (labels_sampled_anchor != labels_sampled_neg).all()



from itertools import chain
from typing import List, Optional, Tuple

import numpy as np
import pytest
import torch
from torch import Tensor

from oml.functional.metrics import calc_retrieval_metrics
from oml.utils.misc_torch import pairwise_dist


def cmc_score_count(distances: Tensor, mask_gt: Tensor, topk: int, mask_to_ignore: Optional[Tensor] = None) -> float:
    metrics = calc_retrieval_metrics(
        distances=distances,
        mask_gt=mask_gt,
        mask_to_ignore=mask_to_ignore,
        cmc_top_k=(topk,),
        map_top_k=tuple(),
        precision_top_k=tuple(),
    )
    return metrics["cmc"][topk]


def cmc_score(
    query_embeddings: Tensor,
    gallery_embeddings: Tensor,
    mask_gt: Tensor,
    topk: int,
    mask_to_ignore: Optional[Tensor] = None,
) -> float:
    distances = pairwise_dist(x1=query_embeddings, x2=gallery_embeddings, p=2)
    val = cmc_score_count(distances=distances, mask_gt=mask_gt, topk=topk, mask_to_ignore=mask_to_ignore)
    return val


EPS = 1e-4

TEST_DATA_SIMPLE = (
    # (distance_matrix, mask_gt,  topk, expected_value)
    (torch.tensor([[1, 2], [2, 1]]), torch.tensor([[0, 1], [1, 0]]), 1, 0.0),
    (torch.tensor([[0, 0.5], [0.0, 0.5]]), torch.tensor([[0, 1], [1, 0]]), 1, 0.5),
    (torch.tensor([[0, 0.5], [0.0, 0.5]]), torch.tensor([[0, 1], [1, 0]]), 2, 1),
    (
        torch.tensor([[1, 0.5, 0.2], [2, 3, 4], [0.4, 3, 4]]),
        torch.tensor([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        2,
        1 / 3,
    ),
    (torch.randn((10, 10)), torch.ones((10, 10)), 1, 1),
)

TEST_DATA_LESS_SMALL = (
    (torch.rand((10, 10)) + torch.tril(torch.ones((10, 10))), torch.eye(10), i, i / 10) for i in range(1, 10)
)

TEST_DATA_GREATER_SMALL = (
    (
        torch.rand((10, 10)) + torch.triu(torch.ones((10, 10)), diagonal=1),
        torch.eye(10),
        i,
        i / 10,
    )
    for i in range(1, 10)
)

TEST_DATA_LESS_BIG = (
    (
        torch.rand((100, 100)) + torch.tril(torch.ones((100, 100))),
        torch.eye(100),
        i,
        i / 100,
    )
    for i in range(1, 101, 10)
)


@pytest.mark.parametrize("distance_matrix,mask_gt,topk,expected", TEST_DATA_SIMPLE)
def test_metric_count(distance_matrix: Tensor, mask_gt: Tensor, topk: int, expected: float) -> None:
    """Simple test"""
    out = cmc_score_count(distances=distance_matrix, mask_gt=mask_gt, topk=topk)
    assert np.isclose(out, expected), (out, expected, "xxx")


@pytest.mark.parametrize(
    "distance_matrix,mask_gt,topk,expected",
    chain(TEST_DATA_LESS_SMALL, TEST_DATA_LESS_BIG),
)
def test_metric_less(distance_matrix: Tensor, mask_gt: Tensor, topk: int, expected: float) -> None:
    """Simple test"""
    out = cmc_score_count(distances=distance_matrix, mask_gt=mask_gt, topk=topk)
    assert out - EPS <= expected


@pytest.mark.parametrize("distance_matrix,mask_gt,topk,expected", chain(TEST_DATA_GREATER_SMALL))
def test_metric_greater(distance_matrix: Tensor, mask_gt: Tensor, topk: int, expected: float) -> None:
    """Simple test"""
    out = cmc_score_count(distances=distance_matrix, mask_gt=mask_gt, topk=topk)
    assert out + EPS >= expected


@pytest.fixture
def generate_samples_for_cmc_score() -> List[Tuple[float, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]]:
    """
    Generate list of query and gallery data for cmc score testing.
    """
    data = []
    for error_rate in [
        0.05,
        0.1,
        0.15,
        0.2,
        0.25,
    ]:
        # generate params of the datasets
        class_number = np.random.randint(low=2, high=5)
        kq = np.random.randint(low=500, high=1000)
        kg = np.random.randint(low=250, high=500)

        def generate_samples(n_labels, samples_per_label):  # type: ignore
            samples = []  # type: ignore
            labels = []  # type: ignore
            # for each label generate dots that will be close to each other and
            # distanced from samples of other labels
            for i in range(n_labels):
                tmp_samples = np.random.uniform(low=2 * i, high=2 * i + 0.2, size=(samples_per_label,))
                samples = np.concatenate((samples, tmp_samples))
                labels = np.concatenate((labels, [i] * samples_per_label))
            return samples.reshape((-1, 1)), labels  # type: ignore

        query_embs, query_labels = generate_samples(n_labels=class_number, samples_per_label=kq)

        gallery_embs, gallery_labels = generate_samples(n_labels=class_number, samples_per_label=kg)

        # spoil generated gallery dataset: for each sample from data change
        # label to any other one with probability error_rate
        def confuse_labels(labels, error_rate):  # type: ignore
            unique_labels = set(labels)
            size = len(labels)
            for i in range(size):
                if np.random.binomial(n=1, p=error_rate, size=1)[0]:
                    labels[i] = np.random.choice(list(unique_labels - {labels[i]}))
            return labels

        gallery_labels = confuse_labels(gallery_labels, error_rate=error_rate)

        query_embs = torch.tensor(query_embs)
        gallery_embs = torch.tensor(gallery_embs)
        query_labels = torch.tensor(query_labels, dtype=torch.long)
        gallery_labels = torch.tensor(gallery_labels, dtype=torch.long)

        data.append((error_rate, query_embs, query_labels, gallery_embs, gallery_labels))
    return data


def test_cmc_score_with_samples(generate_samples_for_cmc_score) -> None:  # type: ignore
    """
    Count cmc score callback for sets of well-separated data clusters labeled
    with error_rate probability mistake.
    """
    for (
        error_rate,
        query_embs,
        query_labels,
        gallery_embs,
        gallery_labels,
    ) in generate_samples_for_cmc_score:
        true_cmc_01 = 1 - error_rate
        mask_gt = (query_labels.reshape((-1, 1)) == gallery_labels).to(torch.bool)
        cmc = cmc_score(
            query_embeddings=query_embs,
            gallery_embeddings=gallery_embs,
            mask_gt=mask_gt,
            topk=1,
        )
        assert abs(cmc - true_cmc_01) <= 0.05


@pytest.mark.parametrize(
    (
        "query_embeddings",
        "gallery_embeddings",
        "mask_gt",
        "available_samples",
        "topk",
        "expected",
    ),
    (
        (
            torch.tensor([[1, 1, 0, 0], [1, 0, 0, 0], [0, 1, 1, 1], [0, 0, 1, 1]]).float(),
            torch.tensor([[1, 1, 1, 0], [1, 1, 1, 1], [0, 1, 1, 0]]).float(),
            torch.tensor(
                [
                    [True, False, False],
                    [True, False, False],
                    [False, True, True],
                    [False, True, True],
                ]
            ),
            torch.tensor(
                [
                    [False, True, True],
                    [True, True, True],
                    [True, False, True],
                    [True, True, True],
                ]
            ),
            1,
            0.75,
        ),
        (
            torch.tensor([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1]]).float(),
            torch.tensor([[0, 1, 0], [0, 0, 1], [1, 0, 1]]).float(),
            torch.tensor(
                [
                    [False, False, True],
                    [True, False, False],
                    [False, True, False],
                    [False, False, True],
                ]
            ),
            torch.tensor(
                [
                    [True, True, True],
                    [False, True, True],
                    [True, False, True],
                    [True, True, False],
                ]
            ),
            1,
            0.25,
        ),
    ),
)
def test_masked_cmc_score(
    query_embeddings: Tensor,
    gallery_embeddings: Tensor,
    mask_gt: Tensor,
    available_samples: Tensor,
    topk: int,
    expected: float,
) -> None:
    score = cmc_score(
        query_embeddings=query_embeddings,
        gallery_embeddings=gallery_embeddings,
        mask_gt=mask_gt,
        mask_to_ignore=~available_samples,
        topk=topk,
    )
    assert score == expected


@pytest.mark.parametrize(
    (
        "query_embeddings",
        "gallery_embeddings",
        "mask_gt",
        "available_samples",
        "topk",
    ),
    (
        (
            torch.rand(size=(query_size, 32)).float(),
            torch.rand(size=(gallery_size, 32)).float(),
            torch.randint(low=0, high=2, size=(query_size, gallery_size)).bool(),
            torch.ones(size=(query_size, gallery_size)).bool(),
            k,
        )
        for query_size, gallery_size, k in zip(list(range(10, 20)), list(range(25, 35)), list(range(1, 11)))
    ),
)
def test_no_mask_cmc_score(
    query_embeddings: Tensor, gallery_embeddings: Tensor, mask_gt: Tensor, available_samples: Tensor, topk: int
) -> None:
    """
    In this test we just check that masked_cmc_score is equal to cmc_score
    when all the samples are available for for scoring.
    """
    masked_score = cmc_score(
        query_embeddings=query_embeddings,
        gallery_embeddings=gallery_embeddings,
        mask_gt=mask_gt,
        mask_to_ignore=~available_samples,
        topk=topk,
    )
    score = cmc_score(
        query_embeddings=query_embeddings,
        gallery_embeddings=gallery_embeddings,
        mask_gt=mask_gt,
        mask_to_ignore=None,
        topk=topk,
    )
    assert masked_score == score



from argparse import ArgumentParser, Namespace
from random import randint, random, sample
from typing import Any, Dict, List, Tuple

import torch
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.utilities.types import (
    EPOCH_OUTPUT,
    EVAL_DATALOADERS,
    TRAIN_DATALOADERS,
)
from torch import nn
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset

from oml.const import TMP_PATH
from oml.ddp.utils import sync_dicts_ddp
from oml.lightning.callbacks.metric import MetricValCallbackDDP
from oml.lightning.entrypoints.parser import parse_engine_params_from_config
from oml.lightning.modules.module_ddp import ModuleDDP
from oml.losses.triplet import TripletLossWithMiner
from oml.metrics.embeddings import EmbeddingMetricsDDP
from oml.samplers.balance import BalanceSampler
from oml.utils.misc import set_global_seed


def create_pred_and_gt_labels(
    num_labels: int, min_max_instances: Tuple[int, int], err_prob: float
) -> Tuple[List[int], List[int]]:
    gt_labels = []
    pred_labels = []

    all_labels = set(range(num_labels))
    for label in range(num_labels):
        instances = [label] * randint(*min_max_instances)
        gt_labels += instances
        if random() < err_prob:
            instances = instances.copy()
            instances[0] = sample(list(all_labels - {label}), k=1)[0]
        pred_labels += instances

    return pred_labels, gt_labels


class DummyDataset(Dataset):
    input_name = "input_tensors"
    labels_name = "labels"
    item_name = "item"

    def __init__(self, pred_labels: List[int], gt_labels: List[int]):
        self.gt_labels = gt_labels
        self.pred_labels = pred_labels

    def __len__(self) -> int:
        return len(self.gt_labels)

    def __getitem__(self, item: int) -> Dict[str, Any]:
        return {
            self.input_name: self.pred_labels[item] * torch.ones((3, 10, 10)),
            self.labels_name: self.gt_labels[item],
            "is_query": True,
            "is_gallery": True,
            self.item_name: item,
        }


class DummyModule(ModuleDDP):
    def __init__(self, loaders_val: EVAL_DATALOADERS, loaders_train: TRAIN_DATALOADERS):
        super().__init__(loaders_val=loaders_val, loaders_train=loaders_train)
        self.model = nn.Sequential(nn.AvgPool2d((10, 10)), nn.Flatten(), nn.Linear(3, 3, bias=False))
        self.criterion = TripletLossWithMiner(margin=None)

    def validation_step(self, batch: Dict[str, Any], batch_idx: int, *dataset_idx: int) -> Dict[str, Any]:
        embeddings = self.model(batch[DummyDataset.input_name])
        return {**batch, **{"embeddings": embeddings}}

    def training_step(self, batch: Dict[str, Any], batch_idx: int) -> Dict[str, Any]:
        embeddings = self.model(batch[DummyDataset.input_name])
        loss = self.criterion(embeddings, batch[DummyDataset.labels_name])
        batch["loss"] = loss
        return batch

    def training_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        self.check_outputs_of_epoch(outputs)

    def validation_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        self.check_outputs_of_epoch(outputs)

    def check_outputs_of_epoch(self, outputs: EPOCH_OUTPUT) -> None:
        # Check point 1 of motivation
        world_size = self.trainer.world_size
        output_batches = [tuple(out[DummyDataset.item_name].tolist()) for out in outputs]
        output_batches_synced = sync_dicts_ddp({"batches": output_batches}, world_size)["batches"]

        assert len(output_batches_synced) == len(output_batches) * world_size
        max_num_not_unique_batches = world_size - 1
        assert len(output_batches_synced) - len(set(output_batches_synced)) <= max_num_not_unique_batches

    def configure_optimizers(self) -> Any:
        return Adam(params=self.parameters(), lr=0.5)


class MetricValCallbackWithSaving(MetricValCallbackDDP):
    """
    We add saving of metrics for later comparison
    """

    save_path_pattern = str(TMP_PATH / "devices_{devices}_batch_size_{batch_size}_num_labels_{num_labels}.pth")

    def __init__(self, devices: int, batch_size: int, num_labels: int, *args: Any, **kwargs: Any):
        super(MetricValCallbackWithSaving, self).__init__(*args, **kwargs)
        self.devices = devices
        self.batch_size = batch_size
        self.num_labels = num_labels

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        ret = super().on_validation_epoch_end(trainer, pl_module)
        save_path = self.save_path_pattern.format(
            devices=self.devices, batch_size=self.batch_size, num_labels=self.num_labels
        )
        torch.save(self.metric.metrics, save_path)  # type: ignore
        return ret


def experiment(args: Namespace) -> None:
    set_global_seed(1)

    devices = args.devices
    max_epochs = args.max_epochs
    num_labels = args.num_labels
    batch_size = args.batch_size

    pred_labels, gt_labels = create_pred_and_gt_labels(num_labels=num_labels, min_max_instances=(3, 5), err_prob=0.3)

    val_dataset = DummyDataset(gt_labels=gt_labels, pred_labels=pred_labels)
    val_dataloader = DataLoader(dataset=val_dataset, batch_size=batch_size)

    train_dataset = DummyDataset(gt_labels=gt_labels, pred_labels=gt_labels)
    batch_sampler = BalanceSampler(labels=gt_labels, n_labels=2, n_instances=2)
    train_dataloader = DataLoader(dataset=train_dataset, batch_sampler=batch_sampler)

    emb_metrics = EmbeddingMetricsDDP(cmc_top_k=(5, 10), precision_top_k=(5, 10), map_top_k=(5, 10))
    val_callback = MetricValCallbackWithSaving(
        metric=emb_metrics, devices=devices, num_labels=num_labels, batch_size=batch_size
    )

    trainer_engine_params = parse_engine_params_from_config({"accelerator": "cpu", "devices": devices})

    pl_model = DummyModule(loaders_val=val_dataloader, loaders_train=train_dataloader)

    trainer = Trainer(
        callbacks=[val_callback],
        max_epochs=max_epochs,
        logger=False,
        enable_checkpointing=False,
        enable_model_summary=False,
        num_sanity_val_steps=0,
        **trainer_engine_params,
    )
    trainer.fit(model=pl_model)


def get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("--devices", type=int)
    parser.add_argument("--max_epochs", type=int)
    parser.add_argument("--num_labels", type=int)
    parser.add_argument("--batch_size", type=int)
    return parser


if __name__ == "__main__":
    experiment(get_parser().parse_args())



from typing import Any

import numpy as np
import pytest
import torch

from oml.metrics.accumulation import Accumulator


def get_data(n: int) -> Any:
    data = {"numpy": np.arange(n), "torch": torch.randn(n, 3), "list": list(range(n))}

    ii = list(range(0, n + 10, 10))
    data_as_batches = []
    for ia, ib in zip(ii, ii[1:]):
        batch = dict()
        for key in data.keys():
            batch[key] = data[key][ia:ib]
        data_as_batches.append(batch)

    return data, data_as_batches


@pytest.mark.parametrize("n", [1, 45, 20])
def test_accumulator(n: int) -> None:
    data, data_as_batches = get_data(n)

    acc = Accumulator(keys_to_accumulate=list(data.keys()))

    for _ in range(3):

        acc.refresh(num_samples=n)
        for batch in data_as_batches:
            acc.update_data(batch)

        assert all((acc.storage[k] == data[k]).all() for k in ["numpy", "torch"])
        assert all((acc.storage[k] == data[k]) for k in ["list"])
        assert acc.collected_samples == n



from typing import Any, Callable, Tuple

from torch.distributed import init_process_group
from torch.multiprocessing import spawn

from oml.const import TMP_PATH
from oml.utils.misc import set_global_seed


def init_ddp(rank: int, world_size: int) -> None:
    if world_size == 0:
        pass
    else:
        init_process_group("gloo", rank=rank, world_size=world_size, init_method=f"file://{TMP_PATH / 'ddp'}")
    set_global_seed(1)


def run_in_ddp(world_size: int, fn: Callable, args: Tuple[Any, ...] = ()) -> None:  # type: ignore
    if world_size == 0:
        return fn(0, world_size, *args)
    # note, 'spawn' automatically passes 'rank' as first argument for 'fn'
    spawn(fn, args=(world_size, *args), nprocs=world_size, join=True)



import cv2
import numpy as np
import pytest

from oml.const import TMP_PATH
from oml.utils.images.images import imread_cv2, imread_pillow


@pytest.mark.parametrize("img_format, no_compression", [("png", True), ("png", False), ("jpg", False)])
@pytest.mark.parametrize("num_channels", [3, 4])
def test_readers(img_format: str, num_channels: int, no_compression: bool) -> None:
    shape_hw = (333, 257)
    dummy_image = np.random.randint(0, 255, (*shape_hw, num_channels), dtype=np.uint8)

    fname_image = str(TMP_PATH / f"img_test_readers.{img_format}")

    if img_format == "png" and no_compression:
        cv2.imwrite(fname_image, cv2.cvtColor(dummy_image, cv2.COLOR_RGBA2BGRA), [cv2.IMWRITE_PNG_COMPRESSION, 0])
    else:
        cv2.imwrite(fname_image, cv2.cvtColor(dummy_image, cv2.COLOR_RGBA2BGRA))

    image_cv2_from_path = imread_cv2(fname_image)
    image_pil_from_path = np.array(imread_pillow(fname_image))

    with open(fname_image, "rb") as fin:
        image_bytes = fin.read()

    image_cv2_from_bytes = imread_cv2(image_bytes)
    image_pil_from_bytes = np.array(imread_pillow(image_bytes))

    for image in [image_cv2_from_path, image_pil_from_path, image_cv2_from_bytes, image_pil_from_bytes]:
        assert image.shape == (*shape_hw, 3)
        if no_compression:
            assert np.array_equal(image, dummy_image[:, :, :3]), (image == dummy_image[:, :, :3]).mean()
        else:
            assert np.array_equal(image, image_cv2_from_path), (image == image_cv2_from_path).mean()



# type: ignore
from typing import Any

import pytest
from omegaconf import OmegaConf
from torch import nn
from torch.optim import Optimizer

from oml.const import CONFIGS_PATH
from oml.registry.losses import LOSSES_REGISTRY, get_criterion
from oml.registry.miners import MINERS_REGISTRY, get_miner
from oml.registry.models import MODELS_REGISTRY, get_extractor
from oml.registry.optimizers import (
    OPTIMIZERS_REGISTRY,
    get_optimizer,
    get_optimizer_by_cfg,
)
from oml.registry.samplers import SAMPLERS_REGISTRY, get_sampler
from oml.registry.schedulers import SCHEDULERS_REGISTRY, get_scheduler
from oml.registry.transforms import TRANSFORMS_REGISTRY, get_transforms
from oml.utils.misc import dictconfig_to_dict


def get_sampler_kwargs_runtime() -> Any:
    return {"label2category": {0: 0, 1: 0, 2: 1, 3: 1}, "labels": [0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3]}


def get_params() -> Any:
    return list(nn.Linear(3, 6).parameters())


def get_opt() -> Optimizer:
    return get_optimizer_by_cfg({"name": "sgd", "args": {"lr": 0.001, "params": get_params()}})


@pytest.mark.parametrize(
    "folder_name,registry,factory_fun,runtime_args",
    [
        ("model", MODELS_REGISTRY, get_extractor, None),
        ("criterion", LOSSES_REGISTRY, get_criterion, None),
        ("miner", MINERS_REGISTRY, get_miner, None),
        ("optimizer", OPTIMIZERS_REGISTRY, get_optimizer, {"params": get_params()}),
        ("sampler", SAMPLERS_REGISTRY, get_sampler, get_sampler_kwargs_runtime()),
        ("scheduler", SCHEDULERS_REGISTRY, get_scheduler, {"optimizer": get_opt()}),
        ("transforms", TRANSFORMS_REGISTRY, get_transforms, None),
    ],
)
def test_registry(folder_name, registry, factory_fun, runtime_args) -> None:
    for obj_name in registry.keys():
        cfg = dictconfig_to_dict(OmegaConf.load(CONFIGS_PATH / folder_name / f"{obj_name}.yaml"))
        args = cfg["args"]

        # this case is special since only 2 schedulers have "lr_lambda" param which is not in defaults
        if (folder_name == "scheduler") and (obj_name == "lambda" or obj_name == "multiplicative"):
            args["lr_lambda"] = lambda epoch: 0.9

        if runtime_args is not None:
            args = dict(**args, **runtime_args)

        factory_fun(cfg["name"], **args)

    assert True
